# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel
#
# Authors: Tomasz Kaczynski <tomasz.kaczynski@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.decision_engine.scoring import scoring_factory
from watcher.tests import base


class TestScoringFactory(base.TestCase):

    def test_get_scoring_engine(self):
        scorer = scoring_factory.get_scoring_engine('dummy_scorer')
        self.assertEqual('dummy_scorer', scorer.get_name())

        scorer = scoring_factory.get_scoring_engine('dummy_min_scorer')
        self.assertEqual('dummy_min_scorer', scorer.get_name())

        scorer = scoring_factory.get_scoring_engine('dummy_max_scorer')
        self.assertEqual('dummy_max_scorer', scorer.get_name())

        scorer = scoring_factory.get_scoring_engine('dummy_avg_scorer')
        self.assertEqual('dummy_avg_scorer', scorer.get_name())

        self.assertRaises(
            KeyError,
            scoring_factory.get_scoring_engine,
            'non_existing_scorer')

    def test_get_scoring_engine_list(self):
        scoring_engines = scoring_factory.get_scoring_engine_list()

        engine_names = {'dummy_scorer', 'dummy_min_scorer',
                        'dummy_max_scorer', 'dummy_avg_scorer'}

        for scorer in scoring_engines:
            self.assertIn(scorer.get_name(), engine_names)



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock


from watcher.common import utils
from watcher.decision_engine.model.collector import base
from watcher.decision_engine.model import element
from watcher.decision_engine.model import model_root as modelroot


volume_uuid_mapping = {
    "volume_0": "5028b1eb-8749-48ae-a42c-5bdd1323976f",
    "volume_1": "74454247-a064-4b34-8f43-89337987720e",
    "volume_2": "a16c811e-2521-4fd3-8779-6a94ccb3be73",
    "volume_3": "37856b95-5be4-4864-8a49-c83f55c66780",
}


class FakerModelCollector(base.BaseClusterDataModelCollector):

    def __init__(self, config=None, osc=None, audit_scope=None):
        if config is None:
            config = mock.Mock(period=777)
        super(FakerModelCollector, self).__init__(config)

    @property
    def notification_endpoints(self):
        return []

    def load_data(self, filename):
        cwd = os.path.abspath(os.path.dirname(__file__))
        data_folder = os.path.join(cwd, "data")

        with open(os.path.join(data_folder, filename), 'rb') as xml_file:
            xml_data = xml_file.read()

        return xml_data

    def load_model(self, filename):
        return modelroot.ModelRoot.from_xml(self.load_data(filename))

    def get_audit_scope_handler(self, audit_scope):
        return None

    def execute(self):
        return self._cluster_data_model or self.build_scenario_1()

    def build_scenario_1(self):
        instances = []

        model = modelroot.ModelRoot()
        # number of nodes
        node_count = 5
        # number max of instance per node
        node_instance_count = 7
        # total number of virtual machine
        instance_count = (node_count * node_instance_count)

        for id_ in range(0, node_count):
            node_uuid = "Node_{0}".format(id_)
            hostname = "hostname_{0}".format(id_)
            node_attributes = {
                "id": id_,
                "uuid": node_uuid,
                "hostname": hostname,
                "memory": 132,
                "memory_mb_reserved": 0,
                "memory_ratio": 1,
                "disk": 250,
                "disk_capacity": 250,
                "disk_gb_reserved": 0,
                "disk_ratio": 1,
                "vcpus": 40,
                "vcpu_reserved": 0,
                "vcpu_ratio": 1,
            }
            node = element.ComputeNode(**node_attributes)
            model.add_node(node)

        for i in range(0, instance_count):
            instance_uuid = "INSTANCE_{0}".format(i)
            if instance_uuid == "INSTANCE_1":
                project_id = "26F03131-32CB-4697-9D61-9123F87A8147"
            elif instance_uuid == "INSTANCE_2":
                project_id = "109F7909-0607-4712-B32C-5CC6D49D2F15"
            else:
                project_id = "91FFFE30-78A0-4152-ACD2-8310FF274DC9"
            instance_attributes = {
                "uuid": instance_uuid,
                "name": instance_uuid,
                "memory": 2,
                "disk": 20,
                "disk_capacity": 20,
                "vcpus": 10,
                "metadata":
                    '{"optimize": true,"top": "floor","nested": {"x": "y"}}',
                "project_id": project_id
            }

            instance = element.Instance(**instance_attributes)
            instances.append(instance)
            model.add_instance(instance)

        mappings = [
            ("INSTANCE_0", "Node_0"),
            ("INSTANCE_1", "Node_0"),
            ("INSTANCE_2", "Node_1"),
            ("INSTANCE_3", "Node_2"),
            ("INSTANCE_4", "Node_2"),
            ("INSTANCE_5", "Node_2"),
            ("INSTANCE_6", "Node_3"),
            ("INSTANCE_7", "Node_4"),
        ]
        for instance_uuid, node_uuid in mappings:
            model.map_instance(
                model.get_instance_by_uuid(instance_uuid),
                model.get_node_by_uuid(node_uuid),
            )

        return model

    def generate_scenario_1(self):
        return self.load_model('scenario_1.xml')

    def generate_scenario_1_with_1_node_unavailable(self):
        return self.load_model('scenario_1_with_1_node_unavailable.xml')

    def generate_scenario_1_with_all_nodes_disable(self):
        return self.load_model('scenario_1_with_all_nodes_disable.xml')

    def generate_scenario_1_with_all_instances_exclude(self):
        return self.load_model('scenario_1_with_all_instances_exclude.xml')

    def generate_scenario_3_with_2_nodes(self):
        return self.load_model('scenario_3_with_2_nodes.xml')

    def generate_scenario_4_with_1_node_no_instance(self):
        return self.load_model('scenario_4_with_1_node_no_instance.xml')

    def generate_scenario_5_with_instance_disk_0(self):
        return self.load_model('scenario_5_with_instance_disk_0.xml')

    def generate_scenario_6_with_2_nodes(self):
        return self.load_model('scenario_6_with_2_nodes.xml')

    def generate_scenario_7_with_2_nodes(self):
        return self.load_model('scenario_7_with_2_nodes.xml')

    def generate_scenario_8_with_4_nodes(self):
        return self.load_model('scenario_8_with_4_nodes.xml')

    def generate_scenario_9_with_3_active_plus_1_disabled_nodes(self):
        return self.load_model(
            'scenario_9_with_3_active_plus_1_disabled_nodes.xml')

    def generate_scenario_10(self):
        return self.load_model('scenario_10.xml')


class FakerStorageModelCollector(base.BaseClusterDataModelCollector):

    def __init__(self, config=None, osc=None, audit_scope=None):
        if config is None:
            config = mock.Mock(period=777)
        super(FakerStorageModelCollector, self).__init__(config)

    @property
    def notification_endpoints(self):
        return []

    def load_data(self, filename):
        cwd = os.path.abspath(os.path.dirname(__file__))
        data_folder = os.path.join(cwd, "data")

        with open(os.path.join(data_folder, filename), 'rb') as xml_file:
            xml_data = xml_file.read()

        return xml_data

    def load_model(self, filename):
        return modelroot.StorageModelRoot.from_xml(self.load_data(filename))

    def get_audit_scope_handler(self, audit_scope):
        return None

    def execute(self):
        return self._cluster_data_model or self.build_scenario_1()

    def build_scenario_1(self):

        model = modelroot.StorageModelRoot()
        # number of nodes
        node_count = 2
        # number of pools per node
        pool_count = 2
        # number of volumes
        volume_count = 9

        for i in range(0, node_count):
            host = "host_{0}@backend_{0}".format(i)
            zone = "zone_{0}".format(i)
            volume_type = ["type_{0}".format(i)]
            node_attributes = {
                "host": host,
                "zone": zone,
                "status": 'enabled',
                "state": 'up',
                "volume_type": volume_type,
            }
            node = element.StorageNode(**node_attributes)
            model.add_node(node)

            for j in range(0, pool_count):
                name = "host_{0}@backend_{0}#pool_{1}".format(i, j)
                pool_attributes = {
                    "name": name,
                    "total_volumes": 2,
                    "total_capacity_gb": 500,
                    "free_capacity_gb": 420,
                    "provisioned_capacity_gb": 80,
                    "allocated_capacity_gb": 80,
                    "virtual_free": 420,
                }
                pool = element.Pool(**pool_attributes)
                model.add_pool(pool)

        mappings = [
            ("host_0@backend_0#pool_0", "host_0@backend_0"),
            ("host_0@backend_0#pool_1", "host_0@backend_0"),
            ("host_1@backend_1#pool_0", "host_1@backend_1"),
            ("host_1@backend_1#pool_1", "host_1@backend_1"),
        ]

        for pool_name, node_name in mappings:
            model.map_pool(
                model.get_pool_by_pool_name(pool_name),
                model.get_node_by_name(node_name),
            )

        volume_uuid_mapping = [
            "5028b1eb-8749-48ae-a42c-5bdd1323976f",
            "74454247-a064-4b34-8f43-89337987720e",
            "a16c811e-2521-4fd3-8779-6a94ccb3be73",
            "37856b95-5be4-4864-8a49-c83f55c66780",
            "694f8fb1-df96-46be-b67d-49f2c14a495e",
            "66b094b0-8fc3-4a94-913f-a5f9312b11a5",
            "e9013810-4b4c-4b94-a056-4c36702d51a3",
            "07976191-6a57-4c35-9f3c-55b3b5ecd6d5",
            "4d1c952d-95d0-4aac-82aa-c3cb509af9f3",
        ]

        for k in range(volume_count):
            uuid = volume_uuid_mapping[k]
            name = "name_{0}".format(k)
            volume_attributes = {
                "size": 40,
                "status": "in-use",
                "uuid": uuid,
                "attachments":
                    '[{"server_id": "server","attachment_id": "attachment"}]',
                "name": name,
                "multiattach": 'True',
                "snapshot_id": uuid,
                "project_id": "91FFFE30-78A0-4152-ACD2-8310FF274DC9",
                "metadata": '{"readonly": false,"attached_mode": "rw"}',
                "bootable": 'False'
            }
            volume = element.Volume(**volume_attributes)
            model.add_volume(volume)

        mappings = [
            (volume_uuid_mapping[0], "host_0@backend_0#pool_0"),
            (volume_uuid_mapping[1], "host_0@backend_0#pool_0"),
            (volume_uuid_mapping[2], "host_0@backend_0#pool_1"),
            (volume_uuid_mapping[3], "host_0@backend_0#pool_1"),
            (volume_uuid_mapping[4], "host_1@backend_1#pool_0"),
            (volume_uuid_mapping[5], "host_1@backend_1#pool_0"),
            (volume_uuid_mapping[6], "host_1@backend_1#pool_1"),
            (volume_uuid_mapping[7], "host_1@backend_1#pool_1"),
        ]

        for volume_uuid, pool_name in mappings:
            model.map_volume(
                model.get_volume_by_uuid(volume_uuid),
                model.get_pool_by_pool_name(pool_name),
            )

        return model

    def generate_scenario_1(self):
        return self.load_model('storage_scenario_1.xml')


class FakerBaremetalModelCollector(base.BaseClusterDataModelCollector):

    def __init__(self, config=None, osc=None):
        if config is None:
            config = mock.Mock(period=777)
        super(FakerBaremetalModelCollector, self).__init__(config)

    @property
    def notification_endpoints(self):
        return []

    def get_audit_scope_handler(self, audit_scope):
        return None

    def load_data(self, filename):
        cwd = os.path.abspath(os.path.dirname(__file__))
        data_folder = os.path.join(cwd, "data")

        with open(os.path.join(data_folder, filename), 'rb') as xml_file:
            xml_data = xml_file.read()

        return xml_data

    def load_model(self, filename):
        return modelroot.BaremetalModelRoot.from_xml(self.load_data(filename))

    def execute(self):
        return self._cluster_data_model or self.build_scenario_1()

    def build_scenario_1(self):
        model = modelroot.BaremetalModelRoot()
        # number of nodes
        node_count = 2

        for i in range(0, node_count):
            uuid = utils.generate_uuid()
            node_attributes = {
                "uuid": uuid,
                "power_state": "power on",
                "maintenance": "false",
                "maintenance_reason": "null",
                "extra": {"compute_node_id": i}
            }
            node = element.IronicNode(**node_attributes)
            model.add_node(node)

        return model

    def generate_scenario_1(self):
        return self.load_model('ironic_scenario_1.xml')



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
A :ref:`Solution <solution_definition>` is the result of execution of a
:ref:`strategy <strategy_definition>` (i.e., an algorithm).
Each solution is composed of many pieces of information:

- A set of :ref:`actions <action_definition>` generated by the strategy in
  order to achieve the :ref:`goal <goal_definition>` of an associated
  :ref:`audit <audit_definition>`.
- A set of :ref:`efficacy indicators <efficacy_indicator_definition>` as
  defined by the associated goal
- A :ref:`global efficacy <efficacy_definition>` which is computed by the
  associated goal using the aforementioned efficacy indicators.

A :ref:`Solution <solution_definition>` is different from an
:ref:`Action Plan <action_plan_definition>` because it contains the
non-scheduled list of :ref:`Actions <action_definition>` which is produced by a
:ref:`Strategy <strategy_definition>`. In other words, the list of Actions in
a :ref:`Solution <solution_definition>` has not yet been re-ordered by the
:ref:`Watcher Planner <watcher_planner_definition>`.

Note that some algorithms (i.e. :ref:`Strategies <strategy_definition>`) may
generate several :ref:`Solutions <solution_definition>`. This gives rise to the
problem of determining which :ref:`Solution <solution_definition>` should be
applied.

Two approaches to dealing with this can be envisaged:

- **fully automated mode**: only the :ref:`Solution <solution_definition>`
  with the highest ranking (i.e., the highest
  :ref:`Optimization Efficacy <efficacy_definition>`) will be sent to the
  :ref:`Watcher Planner <watcher_planner_definition>` and translated into
  concrete :ref:`Actions <action_definition>`.
- **manual mode**: several :ref:`Solutions <solution_definition>` are proposed
  to the :ref:`Administrator <administrator_definition>` with a detailed
  measurement of the estimated :ref:`Optimization Efficacy
  <efficacy_definition>` and he/she decides which one will be launched.
"""

import abc

from watcher.decision_engine.solution import efficacy


class BaseSolution(object, metaclass=abc.ABCMeta):
    def __init__(self, goal, strategy):
        """Base Solution constructor

        :param goal: Goal associated to this solution
        :type goal: :py:class:`~.base.Goal` instance
        :param strategy: Strategy associated to this solution
        :type strategy: :py:class:`~.BaseStrategy` instance
        """
        self.goal = goal
        self._strategy = strategy
        self.origin = None
        self.model = None
        self.efficacy = efficacy.Efficacy(self.goal, self.strategy)

    @property
    def global_efficacy(self):
        return self.efficacy.global_efficacy

    @property
    def efficacy_indicators(self):
        return self.efficacy.indicators

    @property
    def strategy(self):
        return self._strategy

    def compute_global_efficacy(self):
        """Compute the global efficacy given a map of efficacy indicators"""
        self.efficacy.compute_global_efficacy()

    def set_efficacy_indicators(self, **indicators_map):
        """Set the efficacy indicators mapping (no validation)

        :param indicators_map: mapping between the indicator name and its value
        :type indicators_map: dict {`str`: `object`}
        """
        self.efficacy.set_efficacy_indicators(**indicators_map)

    @abc.abstractmethod
    def add_action(self, action_type, resource_id, input_parameters=None):
        """Add a new Action in the Solution

        :param action_type: the unique id of an action type defined in
            entry point 'watcher_actions'
        :param resource_id: the unique id of the resource to which the
            `Action` applies.
        :param input_parameters: An array of input parameters provided as
            key-value pairs of strings. Each key-pair contains names and
            values that match what was previously defined in the `Action`
            type schema.
        """
        raise NotImplementedError()

    @abc.abstractproperty
    def actions(self):
        raise NotImplementedError()



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Servionica LLC
#
# Authors: Alexander Chadin <a.chadin@servionica.ru>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import copy
import itertools
import math
import random

import oslo_cache
from oslo_config import cfg
from oslo_log import log
import oslo_utils

from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.model import element
from watcher.decision_engine.strategy.strategies import base

LOG = log.getLogger(__name__)
CONF = cfg.CONF


def _set_memoize(conf):
    oslo_cache.configure(conf)
    region = oslo_cache.create_region()
    configured_region = oslo_cache.configure_cache_region(conf, region)
    return oslo_cache.core.get_memoization_decorator(conf,
                                                     configured_region,
                                                     'cache')


class WorkloadStabilization(base.WorkloadStabilizationBaseStrategy):
    """Workload Stabilization control using live migration

    This is workload stabilization strategy based on standard deviation
    algorithm. The goal is to determine if there is an overload in a cluster
    and respond to it by migrating VMs to stabilize the cluster.

    This strategy has been tested in a small (32 nodes) cluster.

    It assumes that live migrations are possible in your cluster.
    """

    MEMOIZE = _set_memoize(CONF)

    DATASOURCE_METRICS = ['host_cpu_usage', 'instance_cpu_usage',
                          'instance_ram_usage', 'host_ram_usage']

    def __init__(self, config, osc=None):
        """Workload Stabilization control using live migration

        :param config: A mapping containing the configuration of this strategy
        :type config: :py:class:`~.Struct` instance
        :param osc: :py:class:`~.OpenStackClients` instance
        """
        super(WorkloadStabilization, self).__init__(config, osc)
        self.weights = None
        self.metrics = None
        self.thresholds = None
        self.host_choice = None
        self.instance_metrics = None
        self.retry_count = None
        self.periods = None
        self.aggregation_method = None
        self.sd_before_audit = 0
        self.sd_after_audit = 0
        self.instance_migrations_count = 0
        self.instances_count = 0

    @classmethod
    def get_name(cls):
        return "workload_stabilization"

    @classmethod
    def get_display_name(cls):
        return _("Workload stabilization")

    @classmethod
    def get_translatable_display_name(cls):
        return "Workload stabilization"

    @property
    def granularity(self):
        return self.input_parameters.get('granularity', 300)

    @classmethod
    def get_schema(cls):
        return {
            "properties": {
                "metrics": {
                    "description": "Metrics used as rates of cluster loads.",
                    "type": "array",
                    "items": {
                        "type": "string",
                        "enum": ["instance_cpu_usage", "instance_ram_usage"]
                    },
                    "default": ["instance_cpu_usage"]
                },
                "thresholds": {
                    "description": "Dict where key is a metric and value "
                                   "is a trigger value.",
                    "type": "object",
                    "properties": {
                        "instance_cpu_usage": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        },
                        "instance_ram_usage": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    },
                    "default": {"instance_cpu_usage": 0.1,
                                "instance_ram_usage": 0.1}
                },
                "weights": {
                    "description": "These weights used to calculate "
                                   "common standard deviation. Name of weight"
                                   " contains meter name and _weight suffix.",
                    "type": "object",
                    "properties": {
                        "instance_cpu_usage_weight": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        },
                        "instance_ram_usage_weight": {
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1
                        }
                    },
                    "default": {"instance_cpu_usage_weight": 1.0,
                                "instance_ram_usage_weight": 1.0}
                },
                "instance_metrics": {
                    "description": "Mapping to get hardware statistics using"
                                   " instance metrics",
                    "type": "object",
                    "default": {"instance_cpu_usage": "host_cpu_usage",
                                "instance_ram_usage": "host_ram_usage"}
                },
                "host_choice": {
                    "description": "Method of host's choice. There are cycle,"
                                   " retry and fullsearch methods. "
                                   "Cycle will iterate hosts in cycle. "
                                   "Retry will get some hosts random "
                                   "(count defined in retry_count option). "
                                   "Fullsearch will return each host "
                                   "from list.",
                    "type": "string",
                    "default": "retry"
                },
                "retry_count": {
                    "description": "Count of random returned hosts",
                    "type": "number",
                    "minimum": 1,
                    "default": 1
                },
                "periods": {
                    "description": "These periods are used to get statistic "
                                   "aggregation for instance and host "
                                   "metrics. The period is simply a repeating"
                                   " interval of time into which the samples"
                                   " are grouped for aggregation. Watcher "
                                   "uses only the last period of all received"
                                   " ones.",
                    "type": "object",
                    "properties": {
                        "instance": {
                            "type": "integer",
                            "minimum": 0
                        },
                        "compute_node": {
                            "type": "integer",
                            "minimum": 0
                        },
                        "node": {
                            "type": "integer",
                            # node is deprecated
                            "minimum": 0,
                            "default": 0
                        },
                    },
                    "default": {
                        "instance": 720,
                        "compute_node": 600,
                        # node is deprecated
                        "node": 0,
                    }
                },
                "aggregation_method": {
                    "description": "Function used to aggregate multiple "
                                   "measures into an aggregate. For example, "
                                   "the min aggregation method will aggregate "
                                   "the values of different measures to the "
                                   "minimum value of all the measures in the "
                                   "time range.",
                    "type": "object",
                    "properties": {
                        "instance": {
                            "type": "string",
                            "default": 'mean'
                        },
                        "compute_node": {
                            "type": "string",
                            "default": 'mean'
                        },
                        # node is deprecated
                        "node": {
                            "type": "string",
                            "default": ''
                        },
                    },
                    "default": {
                        "instance": 'mean',
                        "compute_node": 'mean',
                        # node is deprecated
                        "node": '',
                    }
                },
                "granularity": {
                    "description": "The time between two measures in an "
                                   "aggregated timeseries of a metric.",
                    "type": "number",
                    "minimum": 0,
                    "default": 300
                },
            }
        }

    def transform_instance_cpu(self, instance_load, host_vcpus):
        """Transform instance cpu utilization to overall host cpu utilization.

        :param instance_load: dict that contains instance uuid and
            utilization info.
        :param host_vcpus: int
        :return: float value
        """
        return (instance_load['instance_cpu_usage'] *
                (instance_load['vcpus'] / float(host_vcpus)))

    @MEMOIZE
    def get_instance_load(self, instance):
        """Gathering instance load through ceilometer/gnocchi statistic.

        :param instance: instance for which statistic is gathered.
        :return: dict
        """
        LOG.debug('Getting load for %s', instance.uuid)
        instance_load = {'uuid': instance.uuid, 'vcpus': instance.vcpus}
        for meter in self.metrics:
            avg_meter = self.datasource_backend.statistic_aggregation(
                instance, 'instance', meter, self.periods['instance'],
                self.aggregation_method['instance'], self.granularity)
            if avg_meter is None:
                LOG.warning(
                    "No values returned by %(resource_id)s "
                    "for %(metric_name)s", dict(
                        resource_id=instance.uuid, metric_name=meter))
                return
            if meter == 'instance_cpu_usage':
                avg_meter /= float(100)
            LOG.debug('Load of %(metric)s for %(instance)s is %(value)s',
                      {'metric': meter,
                       'instance': instance.uuid,
                       'value': avg_meter})
            instance_load[meter] = avg_meter
        return instance_load

    def normalize_hosts_load(self, hosts):
        normalized_hosts = copy.deepcopy(hosts)
        for host in normalized_hosts:
            if 'instance_ram_usage' in normalized_hosts[host]:
                node = self.compute_model.get_node_by_uuid(host)
                normalized_hosts[host]['instance_ram_usage'] \
                    /= float(node.memory)

        return normalized_hosts

    def get_available_nodes(self):
        nodes = self.compute_model.get_all_compute_nodes().items()
        return {node_uuid: node for node_uuid, node in nodes
                if node.state == element.ServiceState.ONLINE.value and
                node.status == element.ServiceState.ENABLED.value}

    def get_hosts_load(self):
        """Get load of every available host by gathering instances load"""
        hosts_load = {}
        for node_id, node in self.get_available_nodes().items():
            hosts_load[node_id] = {}
            hosts_load[node_id]['vcpus'] = node.vcpus
            LOG.debug('Getting load for %s', node_id)
            for metric in self.metrics:
                avg_meter = None
                meter_name = self.instance_metrics[metric]
                avg_meter = self.datasource_backend.statistic_aggregation(
                    node, 'compute_node', self.instance_metrics[metric],
                    self.periods['compute_node'],
                    self.aggregation_method['compute_node'], self.granularity)
                if avg_meter is None:
                    LOG.warning('No values returned by node %s for %s',
                                node_id, meter_name)
                    del hosts_load[node_id]
                    break
                else:
                    if meter_name == 'host_ram_usage':
                        avg_meter /= oslo_utils.units.Ki
                    if meter_name == 'host_cpu_usage':
                        avg_meter /= 100
                LOG.debug('Load of %(metric)s for %(node)s is %(value)s',
                          {'metric': metric,
                           'node': node_id,
                           'value': avg_meter})
                hosts_load[node_id][metric] = avg_meter
        return hosts_load

    def get_sd(self, hosts, meter_name):
        """Get standard deviation among hosts by specified meter"""
        mean = 0
        variation = 0
        num_hosts = len(hosts)
        if num_hosts == 0:
            return 0
        for host_id in hosts:
            mean += hosts[host_id][meter_name]
        mean /= num_hosts
        for host_id in hosts:
            variation += (hosts[host_id][meter_name] - mean) ** 2
        variation /= num_hosts
        sd = math.sqrt(variation)
        return sd

    def calculate_weighted_sd(self, sd_case):
        """Calculate common standard deviation among meters on host"""
        weighted_sd = 0
        for metric, value in zip(self.metrics, sd_case):
            try:
                weighted_sd += value * float(self.weights[metric + '_weight'])
            except KeyError as exc:
                LOG.exception(exc)
                raise exception.WatcherException(
                    _("Incorrect mapping: could not find associated weight"
                      " for %s in weight dict.") % metric)
        return weighted_sd

    def calculate_migration_case(self, hosts, instance, src_node, dst_node):
        """Calculate migration case

        Return list of standard deviation values, that appearing in case of
        migration of instance from source host to destination host
        :param hosts: hosts with their workload
        :param instance: the virtual machine
        :param src_node: the source node
        :param dst_node: the destination node
        :return: list of standard deviation values
        """
        migration_case = []
        new_hosts = copy.deepcopy(hosts)
        instance_load = self.get_instance_load(instance)
        if not instance_load:
            return
        s_host_vcpus = new_hosts[src_node.uuid]['vcpus']
        d_host_vcpus = new_hosts[dst_node.uuid]['vcpus']
        for metric in self.metrics:
            if metric == 'instance_cpu_usage':
                new_hosts[src_node.uuid][metric] -= (
                    self.transform_instance_cpu(instance_load, s_host_vcpus))
                new_hosts[dst_node.uuid][metric] += (
                    self.transform_instance_cpu(instance_load, d_host_vcpus))
            else:
                new_hosts[src_node.uuid][metric] -= instance_load[metric]
                new_hosts[dst_node.uuid][metric] += instance_load[metric]
        normalized_hosts = self.normalize_hosts_load(new_hosts)
        for metric in self.metrics:
            migration_case.append(self.get_sd(normalized_hosts, metric))
        migration_case.append(new_hosts)
        return migration_case

    def get_current_weighted_sd(self, hosts_load):
        """Calculate current weighted sd"""
        current_sd = []
        normalized_load = self.normalize_hosts_load(hosts_load)
        for metric in self.metrics:
            metric_sd = self.get_sd(normalized_load, metric)
            current_sd.append(metric_sd)
        current_sd.append(hosts_load)
        return self.calculate_weighted_sd(current_sd[:-1])

    def simulate_migrations(self, hosts):
        """Make sorted list of pairs instance:dst_host"""
        def yield_nodes(nodes):
            if self.host_choice == 'cycle':
                for i in itertools.cycle(nodes):
                    yield [i]
            if self.host_choice == 'retry':
                while True:
                    yield random.sample(nodes, self.retry_count)
            if self.host_choice == 'fullsearch':
                while True:
                    yield nodes

        instance_host_map = []
        nodes = sorted(list(self.get_available_nodes()))
        current_weighted_sd = self.get_current_weighted_sd(hosts)
        for src_host in nodes:
            src_node = self.compute_model.get_node_by_uuid(src_host)
            c_nodes = copy.copy(nodes)
            c_nodes.remove(src_host)
            node_list = yield_nodes(c_nodes)
            for instance in self.compute_model.get_node_instances(src_node):
                # NOTE: skip exclude instance when migrating
                if instance.watcher_exclude:
                    LOG.debug("Instance is excluded by scope, "
                              "skipped: %s", instance.uuid)
                    continue
                if instance.state not in [element.InstanceState.ACTIVE.value,
                                          element.InstanceState.PAUSED.value]:
                    continue
                min_sd_case = {'value': current_weighted_sd}
                for dst_host in next(node_list):
                    dst_node = self.compute_model.get_node_by_uuid(dst_host)
                    sd_case = self.calculate_migration_case(
                        hosts, instance, src_node, dst_node)
                    if sd_case is None:
                        break

                    weighted_sd = self.calculate_weighted_sd(sd_case[:-1])

                    if weighted_sd < min_sd_case['value']:
                        min_sd_case = {
                            'host': dst_node.uuid, 'value': weighted_sd,
                            's_host': src_node.uuid, 'instance': instance.uuid}
                        instance_host_map.append(min_sd_case)
                if sd_case is None:
                    continue
        return sorted(instance_host_map, key=lambda x: x['value'])

    def check_threshold(self):
        """Check if cluster is needed in balancing"""
        hosts_load = self.get_hosts_load()
        normalized_load = self.normalize_hosts_load(hosts_load)
        for metric in self.metrics:
            metric_sd = self.get_sd(normalized_load, metric)
            LOG.info("Standard deviation for %(metric)s is %(sd)s.",
                     {'metric': metric, 'sd': metric_sd})
            if metric_sd > float(self.thresholds[metric]):
                LOG.info("Standard deviation of %(metric)s exceeds"
                         " appropriate threshold %(threshold)s by %(sd)s.",
                         {'metric': metric,
                          'threshold': float(self.thresholds[metric]),
                          'sd': metric_sd})
                LOG.info("Launching workload optimization...")
                self.sd_before_audit = metric_sd
                return self.simulate_migrations(hosts_load)

    def create_migration_instance(self, mig_instance, mig_source_node,
                                  mig_destination_node):
        """Create migration VM"""
        if self.compute_model.migrate_instance(
                mig_instance, mig_source_node, mig_destination_node):
            self.add_action_migrate(mig_instance, 'live',
                                    mig_source_node,
                                    mig_destination_node)
            self.instance_migrations_count += 1

    def migrate(self, instance_uuid, src_host, dst_host):
        mig_instance = self.compute_model.get_instance_by_uuid(instance_uuid)
        mig_source_node = self.compute_model.get_node_by_uuid(
            src_host)
        mig_destination_node = self.compute_model.get_node_by_uuid(
            dst_host)
        self.create_migration_instance(mig_instance, mig_source_node,
                                       mig_destination_node)

    def fill_solution(self):
        self.solution.model = self.compute_model
        return self.solution

    def pre_execute(self):
        self._pre_execute()
        self.weights = self.input_parameters.weights
        self.metrics = self.input_parameters.metrics
        self.thresholds = self.input_parameters.thresholds
        self.host_choice = self.input_parameters.host_choice
        self.instance_metrics = self.input_parameters.instance_metrics
        self.retry_count = self.input_parameters.retry_count
        self.periods = self.input_parameters.periods
        self.aggregation_method = self.input_parameters.aggregation_method

        # backwards compatibility for node parameter with aggregate.
        if self.aggregation_method['node']:
            LOG.warning('Parameter node has been renamed to compute_node and '
                        'will be removed in next release.')
            self.aggregation_method['compute_node'] = \
                self.aggregation_method['node']

        # backwards compatibility for node parameter with period.
        if self.periods['node'] != 0:
            LOG.warning('Parameter node has been renamed to compute_node and '
                        'will be removed in next release.')
            self.periods['compute_node'] = self.periods['node']

    def do_execute(self, audit=None):
        migration = self.check_threshold()
        if migration:
            hosts_load = self.get_hosts_load()
            min_sd = 1
            balanced = False
            for instance_host in migration:
                instance = self.compute_model.get_instance_by_uuid(
                    instance_host['instance'])
                src_node = self.compute_model.get_node_by_uuid(
                    instance_host['s_host'])
                dst_node = self.compute_model.get_node_by_uuid(
                    instance_host['host'])
                if instance.disk > dst_node.disk:
                    continue
                instance_load = self.calculate_migration_case(
                    hosts_load, instance, src_node, dst_node)
                weighted_sd = self.calculate_weighted_sd(instance_load[:-1])
                if weighted_sd < min_sd:
                    min_sd = weighted_sd
                    hosts_load = instance_load[-1]
                    LOG.info("Migration of %(instance_uuid)s from %(s_host)s "
                             "to %(host)s reduces standard deviation to "
                             "%(min_sd)s.",
                             {'instance_uuid': instance_host['instance'],
                              's_host': instance_host['s_host'],
                              'host': instance_host['host'],
                              'min_sd': min_sd})
                    self.migrate(instance_host['instance'],
                                 instance_host['s_host'],
                                 instance_host['host'])
                    self.sd_after_audit = min_sd

                for metric, value in zip(self.metrics, instance_load[:-1]):
                    if value < float(self.thresholds[metric]):
                        LOG.info("At least one of metrics' values fell "
                                 "below the threshold values. "
                                 "Workload Stabilization has successfully "
                                 "completed optimization process.")
                        balanced = True
                        break
                if balanced:
                    break

    def post_execute(self):
        """Post-execution phase

        This can be used to compute the global efficacy
        """
        self.fill_solution()

        self.solution.set_efficacy_indicators(
            instance_migrations_count=self.instance_migrations_count,
            standard_deviation_before_audit=self.sd_before_audit,
            standard_deviation_after_audit=self.sd_after_audit,
            instances_count=len(self.compute_model.get_all_instances()),
        )

        LOG.debug(self.compute_model.to_string())



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel Corp
#
# Authors: Junjie-Huang <junjie.huang@intel.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_log import log

from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.model import element
from watcher.decision_engine.strategy.strategies import base

LOG = log.getLogger(__name__)


class WorkloadBalance(base.WorkloadStabilizationBaseStrategy):
    """[PoC]Workload balance using live migration

    *Description*

        It is a migration strategy based on the VM workload of physical
        servers. It generates solutions to move a workload whenever a server's
        CPU or RAM utilization % is higher than the specified threshold.
        The VM to be moved should make the host close to average workload
        of all compute nodes.

    *Requirements*

        * Hardware: compute node should use the same physical CPUs/RAMs
        * Software: Ceilometer component ceilometer-agent-compute running
          in each compute node, and Ceilometer API can report such telemetry
          "instance_cpu_usage" and "instance_ram_usage" successfully.
        * You must have at least 2 physical compute nodes to run this strategy.

    *Limitations*

       - This is a proof of concept that is not meant to be used in production
       - We cannot forecast how many servers should be migrated. This is the
         reason why we only plan a single virtual machine migration at a time.
         So it's better to use this algorithm with `CONTINUOUS` audits.
       - It assume that live migrations are possible
    """

    # The meter to report CPU utilization % of VM in ceilometer
    # Unit: %, value range is [0 , 100]

    # The meter to report memory resident of VM in ceilometer
    # Unit: MB

    DATASOURCE_METRICS = ['instance_cpu_usage', 'instance_ram_usage']

    def __init__(self, config, osc=None):
        """Workload balance using live migration

        :param config: A mapping containing the configuration of this strategy
        :type config: :py:class:`~.Struct` instance
        :param osc: :py:class:`~.OpenStackClients` instance
        """
        super(WorkloadBalance, self).__init__(config, osc)
        # the migration plan will be triggered when the CPU or RAM
        # utilization % reaches threshold
        self._meter = None
        self.instance_migrations_count = 0

    @classmethod
    def get_name(cls):
        return "workload_balance"

    @classmethod
    def get_display_name(cls):
        return _("Workload Balance Migration Strategy")

    @classmethod
    def get_translatable_display_name(cls):
        return "Workload Balance Migration Strategy"

    @property
    def granularity(self):
        return self.input_parameters.get('granularity', 300)

    @classmethod
    def get_schema(cls):
        # Mandatory default setting for each element
        return {
            "properties": {
                "metrics": {
                    "description": "Workload balance based on metrics: "
                                   "cpu or ram utilization",
                    "type": "string",
                    "choice": ["instance_cpu_usage", "instance_ram_usage"],
                    "default": "instance_cpu_usage"
                },
                "threshold": {
                    "description": "workload threshold for migration",
                    "type": "number",
                    "default": 25.0
                },
                "period": {
                    "description": "aggregate time period of ceilometer",
                    "type": "number",
                    "default": 300
                },
                "granularity": {
                    "description": "The time between two measures in an "
                                   "aggregated timeseries of a metric.",
                    "type": "number",
                    "default": 300
                },
            },
        }

    def get_available_compute_nodes(self):
        default_node_scope = [element.ServiceState.ENABLED.value]
        return {uuid: cn for uuid, cn in
                self.compute_model.get_all_compute_nodes().items()
                if cn.state == element.ServiceState.ONLINE.value and
                cn.status in default_node_scope}

    def choose_instance_to_migrate(self, hosts, avg_workload, workload_cache):
        """Pick up an active instance to migrate from provided hosts

        :param hosts: the array of dict which contains node object
        :param avg_workload: the average workload value of all nodes
        :param workload_cache: the map contains instance to workload mapping
        """
        for instance_data in hosts:
            source_node = instance_data['compute_node']
            source_instances = self.compute_model.get_node_instances(
                source_node)
            if source_instances:
                delta_workload = instance_data['workload'] - avg_workload
                min_delta = 1000000
                instance_id = None
                for instance in source_instances:
                    try:
                        # NOTE: skip exclude instance when migrating
                        if instance.watcher_exclude:
                            LOG.debug("Instance is excluded by scope, "
                                      "skipped: %s", instance.uuid)
                            continue
                        # select the first active VM to migrate
                        if (instance.state !=
                                element.InstanceState.ACTIVE.value):
                            LOG.debug("Instance not active, skipped: %s",
                                      instance.uuid)
                            continue
                        current_delta = (
                            delta_workload - workload_cache[instance.uuid])
                        if 0 <= current_delta < min_delta:
                            min_delta = current_delta
                            instance_id = instance.uuid
                    except exception.InstanceNotFound:
                        LOG.error("Instance not found; error: %s",
                                  instance_id)
                if instance_id:
                    return (source_node,
                            self.compute_model.get_instance_by_uuid(
                                instance_id))
            else:
                LOG.info("VM not found from compute_node: %s",
                         source_node.uuid)

    def filter_destination_hosts(self, hosts, instance_to_migrate,
                                 avg_workload, workload_cache):
        """Only return hosts with sufficient available resources"""
        required_cores = instance_to_migrate.vcpus
        required_disk = instance_to_migrate.disk
        required_mem = instance_to_migrate.memory

        # filter nodes without enough resource
        destination_hosts = []
        src_instance_workload = workload_cache[instance_to_migrate.uuid]
        for instance_data in hosts:
            host = instance_data['compute_node']
            workload = instance_data['workload']
            # calculate the available resources
            free_res = self.compute_model.get_node_free_resources(host)
            if (free_res['vcpu'] >= required_cores and
                    free_res['memory'] >= required_mem and
                    free_res['disk'] >= required_disk):
                if (self._meter == 'instance_cpu_usage' and
                    ((src_instance_workload + workload) <
                     self.threshold / 100 * host.vcpus)):
                    destination_hosts.append(instance_data)
                if (self._meter == 'instance_ram_usage' and
                    ((src_instance_workload + workload) <
                     self.threshold / 100 * host.memory)):
                    destination_hosts.append(instance_data)

        return destination_hosts

    def group_hosts_by_cpu_or_ram_util(self):
        """Calculate the workloads of each compute_node

        try to find out the nodes which have reached threshold
        and the nodes which are under threshold.
        and also calculate the average workload value of all nodes.
        and also generate the instance workload map.
        """

        nodes = self.get_available_compute_nodes()
        cluster_size = len(nodes)
        overload_hosts = []
        nonoverload_hosts = []
        # total workload of cluster
        cluster_workload = 0.0
        # use workload_cache to store the workload of VMs for reuse purpose
        workload_cache = {}
        for node_id in nodes:
            node = self.compute_model.get_node_by_uuid(node_id)
            instances = self.compute_model.get_node_instances(node)
            node_workload = 0.0
            for instance in instances:
                util = None
                try:
                    util = self.datasource_backend.statistic_aggregation(
                        instance, 'instance', self._meter, self._period,
                        'mean', self._granularity)
                except Exception as exc:
                    LOG.exception(exc)
                    LOG.error("Can not get %s from %s", self._meter,
                              self.datasource_backend.NAME)
                    continue
                if util is None:
                    LOG.debug("Instance (%s): %s is None",
                              instance.uuid, self._meter)
                    continue
                if self._meter == 'instance_cpu_usage':
                    workload_cache[instance.uuid] = (util *
                                                     instance.vcpus / 100)
                else:
                    workload_cache[instance.uuid] = util
                node_workload += workload_cache[instance.uuid]
                LOG.debug("VM (%s): %s %f", instance.uuid, self._meter,
                          util)

            cluster_workload += node_workload
            if self._meter == 'instance_cpu_usage':
                node_util = node_workload / node.vcpus * 100
            else:
                node_util = node_workload / node.memory * 100

            instance_data = {
                'compute_node': node, self._meter: node_util,
                'workload': node_workload}
            if node_util >= self.threshold:
                # mark the node to release resources
                overload_hosts.append(instance_data)
            else:
                nonoverload_hosts.append(instance_data)

        avg_workload = 0
        if cluster_size != 0:
            avg_workload = cluster_workload / cluster_size

        return overload_hosts, nonoverload_hosts, avg_workload, workload_cache

    def pre_execute(self):
        self._pre_execute()
        self.threshold = self.input_parameters.threshold
        self._period = self.input_parameters.period
        self._meter = self.input_parameters.metrics
        self._granularity = self.input_parameters.granularity

    def do_execute(self, audit=None):
        """Strategy execution phase

        This phase is where you should put the main logic of your strategy.
        """
        source_nodes, target_nodes, avg_workload, workload_cache = (
            self.group_hosts_by_cpu_or_ram_util())

        if not source_nodes:
            LOG.debug("No hosts require optimization")
            return self.solution

        if not target_nodes:
            LOG.warning("No hosts current have CPU utilization under %s "
                        "percent, therefore there are no possible target "
                        "hosts for any migration",
                        self.threshold)
            return self.solution

        # choose the server with largest cpu_util
        source_nodes = sorted(source_nodes,
                              reverse=True,
                              key=lambda x: (x[self._meter]))

        instance_to_migrate = self.choose_instance_to_migrate(
            source_nodes, avg_workload, workload_cache)
        if not instance_to_migrate:
            return self.solution
        source_node, instance_src = instance_to_migrate
        # find the hosts that have enough resource for the VM to be migrated
        destination_hosts = self.filter_destination_hosts(
            target_nodes, instance_src, avg_workload, workload_cache)
        # sort the filtered result by workload
        # pick up the lowest one as dest server
        if not destination_hosts:
            # for instance.
            LOG.warning("No proper target host could be found, it might "
                        "be because of there's no enough CPU/Memory/DISK")
            return self.solution
        destination_hosts = sorted(destination_hosts,
                                   key=lambda x: (x[self._meter]))
        # always use the host with lowerest CPU utilization
        mig_destination_node = destination_hosts[0]['compute_node']
        # generate solution to migrate the instance to the dest server,
        if self.compute_model.migrate_instance(
                instance_src, source_node, mig_destination_node):
            self.add_action_migrate(
                instance_src,
                'live',
                source_node,
                mig_destination_node)
            self.instance_migrations_count += 1

    def post_execute(self):
        """Post-execution phase

        This can be used to compute the global efficacy
        """
        self.solution.model = self.compute_model
        self.solution.set_efficacy_indicators(
            instance_migrations_count=self.instance_migrations_count,
            instances_count=len(self.compute_model.get_all_instances())
        )

        LOG.debug(self.compute_model.to_string())



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Intel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Tests for manipulating ScoringEngine via the DB API"""

import freezegun

from watcher.common import exception
from watcher.common import utils as w_utils
from watcher.tests.db import base
from watcher.tests.db import utils


class TestDbScoringEngineFilters(base.DbTestCase):

    FAKE_OLDER_DATE = '2014-01-01T09:52:05.219414'
    FAKE_OLD_DATE = '2015-01-01T09:52:05.219414'
    FAKE_TODAY = '2016-02-24T09:52:05.219414'

    def setUp(self):
        super(TestDbScoringEngineFilters, self).setUp()
        self.context.show_deleted = True
        self._data_setup()

    def _data_setup(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.scoring_engine1 = utils.create_test_scoring_engine(
                id=1, uuid='e8370ede-4f39-11e6-9ffa-08002722cb22',
                name="se-1", description="Scoring Engine 1", metainfo="a1=b1")
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.scoring_engine2 = utils.create_test_scoring_engine(
                id=2, uuid='e8370ede-4f39-11e6-9ffa-08002722cb23',
                name="se-2", description="Scoring Engine 2", metainfo="a2=b2")
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.scoring_engine3 = utils.create_test_scoring_engine(
                id=3, uuid='e8370ede-4f39-11e6-9ffa-08002722cb24',
                name="se-3", description="Scoring Engine 3", metainfo="a3=b3")

    def _soft_delete_scoring_engines(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_scoring_engine(self.scoring_engine1.id)
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.dbapi.soft_delete_scoring_engine(self.scoring_engine2.id)
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.dbapi.soft_delete_scoring_engine(self.scoring_engine3.id)

    def _update_scoring_engines(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.update_scoring_engine(
                self.scoring_engine1.id,
                values={"description": "scoring_engine1"})
        with freezegun.freeze_time(self.FAKE_OLD_DATE):
            self.dbapi.update_scoring_engine(
                self.scoring_engine2.id,
                values={"description": "scoring_engine2"})
        with freezegun.freeze_time(self.FAKE_OLDER_DATE):
            self.dbapi.update_scoring_engine(
                self.scoring_engine3.id,
                values={"description": "scoring_engine3"})

    def test_get_scoring_engine_list_filter_deleted_true(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_scoring_engine(self.scoring_engine1.id)

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted': True})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_deleted_false(self):
        with freezegun.freeze_time(self.FAKE_TODAY):
            self.dbapi.soft_delete_scoring_engine(self.scoring_engine1.id)

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted': False})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_deleted_at_eq(self):
        self._soft_delete_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_deleted_at_lt(self):
        self._soft_delete_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_deleted_at_lte(self):
        self._soft_delete_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_deleted_at_gt(self):
        self._soft_delete_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_deleted_at_gte(self):
        self._soft_delete_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'deleted_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine1['id'], self.scoring_engine2['id']]),
            set([r.id for r in res]))

    # created_at #

    def test_get_scoring_engine_list_filter_created_at_eq(self):
        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'created_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_created_at_lt(self):
        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'created_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_created_at_lte(self):
        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'created_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_created_at_gt(self):
        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'created_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_created_at_gte(self):
        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'created_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine1['id'], self.scoring_engine2['id']]),
            set([r.id for r in res]))

    # updated_at #

    def test_get_scoring_engine_list_filter_updated_at_eq(self):
        self._update_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'updated_at__eq': self.FAKE_TODAY})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_updated_at_lt(self):
        self._update_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'updated_at__lt': self.FAKE_TODAY})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_updated_at_lte(self):
        self._update_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'updated_at__lte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine2['id'], self.scoring_engine3['id']]),
            set([r.id for r in res]))

    def test_get_scoring_engine_list_filter_updated_at_gt(self):
        self._update_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'updated_at__gt': self.FAKE_OLD_DATE})

        self.assertEqual([self.scoring_engine1['id']], [r.id for r in res])

    def test_get_scoring_engine_list_filter_updated_at_gte(self):
        self._update_scoring_engines()

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'updated_at__gte': self.FAKE_OLD_DATE})

        self.assertEqual(
            set([self.scoring_engine1['id'], self.scoring_engine2['id']]),
            set([r.id for r in res]))


class DbScoringEngineTestCase(base.DbTestCase):

    def test_get_scoring_engine_list(self):
        names = []
        for i in range(1, 4):
            scoring_engine = utils.create_test_scoring_engine(
                id=i,
                uuid=w_utils.generate_uuid(),
                name="SE_ID_%s" % i,
                description='My ScoringEngine {0}'.format(i),
                metainfo='a{0}=b{0}'.format(i))
            names.append(str(scoring_engine['name']))
        scoring_engines = self.dbapi.get_scoring_engine_list(self.context)
        scoring_engines_names = [se.name for se in scoring_engines]
        self.assertEqual(sorted(names), sorted(scoring_engines_names))

    def test_get_scoring_engine_list_with_filters(self):
        scoring_engine1 = utils.create_test_scoring_engine(
            id=1,
            uuid=w_utils.generate_uuid(),
            name="SE_ID_1",
            description='ScoringEngine 1',
            metainfo="a1=b1",
        )
        scoring_engine2 = utils.create_test_scoring_engine(
            id=2,
            uuid=w_utils.generate_uuid(),
            name="SE_ID_2",
            description='ScoringEngine 2',
            metainfo="a2=b2",
        )
        scoring_engine3 = utils.create_test_scoring_engine(
            id=3,
            uuid=w_utils.generate_uuid(),
            name="SE_ID_3",
            description='ScoringEngine 3',
            metainfo="a3=b3",
        )

        self.dbapi.soft_delete_scoring_engine(scoring_engine3['uuid'])

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'description': 'ScoringEngine 1'})
        self.assertEqual([scoring_engine1['name']], [r.name for r in res])

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'description': 'ScoringEngine 3'})
        self.assertEqual([], [r.name for r in res])

        res = self.dbapi.get_scoring_engine_list(
            self.context, filters={'description': 'ScoringEngine 2'})
        self.assertEqual([scoring_engine2['name']], [r.name for r in res])

    def test_get_scoring_engine_by_id(self):
        created_scoring_engine = utils.create_test_scoring_engine()
        scoring_engine = self.dbapi.get_scoring_engine_by_id(
            self.context, created_scoring_engine['id'])
        self.assertEqual(scoring_engine.id, created_scoring_engine['id'])

    def test_get_scoring_engine_by_uuid(self):
        created_scoring_engine = utils.create_test_scoring_engine()
        scoring_engine = self.dbapi.get_scoring_engine_by_uuid(
            self.context, created_scoring_engine['uuid'])
        self.assertEqual(scoring_engine.uuid, created_scoring_engine['uuid'])

    def test_get_scoring_engine_by_name(self):
        created_scoring_engine = utils.create_test_scoring_engine()
        scoring_engine = self.dbapi.get_scoring_engine_by_name(
            self.context, created_scoring_engine['name'])
        self.assertEqual(scoring_engine.name, created_scoring_engine['name'])

    def test_get_scoring_engine_that_does_not_exist(self):
        self.assertRaises(exception.ScoringEngineNotFound,
                          self.dbapi.get_scoring_engine_by_id,
                          self.context, 404)

    def test_update_scoring_engine(self):
        scoring_engine = utils.create_test_scoring_engine()
        res = self.dbapi.update_scoring_engine(
            scoring_engine['id'], {'description': 'updated-model'})
        self.assertEqual('updated-model', res.description)

    def test_update_scoring_engine_id(self):
        scoring_engine = utils.create_test_scoring_engine()
        self.assertRaises(exception.Invalid,
                          self.dbapi.update_scoring_engine,
                          scoring_engine['id'],
                          {'uuid': w_utils.generate_uuid()})

    def test_update_scoring_engine_that_does_not_exist(self):
        self.assertRaises(exception.ScoringEngineNotFound,
                          self.dbapi.update_scoring_engine,
                          404,
                          {'description': ''})

    def test_destroy_scoring_engine(self):
        scoring_engine = utils.create_test_scoring_engine()
        self.dbapi.destroy_scoring_engine(scoring_engine['id'])
        self.assertRaises(exception.ScoringEngineNotFound,
                          self.dbapi.get_scoring_engine_by_id,
                          self.context, scoring_engine['id'])

    def test_destroy_scoring_engine_that_does_not_exist(self):
        self.assertRaises(exception.ScoringEngineNotFound,
                          self.dbapi.destroy_scoring_engine, 404)

    def test_create_scoring_engine_already_exists(self):
        scoring_engine_id = "SE_ID"
        utils.create_test_scoring_engine(name=scoring_engine_id)
        self.assertRaises(exception.ScoringEngineAlreadyExists,
                          utils.create_test_scoring_engine,
                          name=scoring_engine_id)



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 Servionica
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from unittest import mock

from oslo_config import cfg

from watcher.common import clients
from watcher.common import exception
from watcher.decision_engine.datasources import gnocchi as gnocchi_helper
from watcher.tests import base

CONF = cfg.CONF


@mock.patch.object(clients.OpenStackClients, 'gnocchi')
class TestGnocchiHelper(base.BaseTestCase):

    def setUp(self):
        super(TestGnocchiHelper, self).setUp()
        self.osc_mock = mock.Mock()
        self.helper = gnocchi_helper.GnocchiHelper(osc=self.osc_mock)
        stat_agg_patcher = mock.patch.object(
            self.helper, 'statistic_aggregation',
            spec=gnocchi_helper.GnocchiHelper.statistic_aggregation)
        self.mock_aggregation = stat_agg_patcher.start()
        self.addCleanup(stat_agg_patcher.stop)

    def test_gnocchi_statistic_aggregation(self, mock_gnocchi):
        gnocchi = mock.MagicMock()
        expected_result = 5.5

        expected_measures = [["2017-02-02T09:00:00.000000", 360, 5.5]]

        gnocchi.metric.get_measures.return_value = expected_measures
        mock_gnocchi.return_value = gnocchi

        helper = gnocchi_helper.GnocchiHelper()
        result = helper.statistic_aggregation(
            resource=mock.Mock(id='16a86790-327a-45f9-bc82-45839f062fdc'),
            resource_type='instance',
            meter_name='instance_cpu_usage',
            period=300,
            granularity=360,
            aggregate='mean',
        )
        self.assertEqual(expected_result, result)

    def test_gnocchi_statistic_series(self, mock_gnocchi):
        gnocchi = mock.MagicMock()
        expected_result = {
            "2017-02-02T09:00:00.000000": 5.5,
            "2017-02-02T09:03:60.000000": 5.8
        }

        expected_measures = [
            ["2017-02-02T09:00:00.000000", 360, 5.5],
            ["2017-02-02T09:03:60.000000", 360, 5.8]
        ]

        gnocchi.metric.get_measures.return_value = expected_measures
        mock_gnocchi.return_value = gnocchi

        start = datetime(year=2017, month=2, day=2, hour=9, minute=0)
        end = datetime(year=2017, month=2, day=2, hour=9, minute=4)

        helper = gnocchi_helper.GnocchiHelper()
        result = helper.statistic_series(
            resource=mock.Mock(id='16a86790-327a-45f9-bc82-45839f062fdc'),
            resource_type='instance',
            meter_name='instance_cpu_usage',
            start_time=start,
            end_time=end,
            granularity=360,
        )
        self.assertEqual(expected_result, result)

    def test_statistic_aggregation_metric_unavailable(self, mock_gnocchi):
        helper = gnocchi_helper.GnocchiHelper()

        # invalidate instance_cpu_usage in metric map
        original_metric_value = helper.METRIC_MAP.get('instance_cpu_usage')
        helper.METRIC_MAP.update(
            instance_cpu_usage=None
        )

        self.assertRaises(
            exception.MetricNotAvailable, helper.statistic_aggregation,
            resource=mock.Mock(id='16a86790-327a-45f9-bc82-45839f062fdc'),
            resource_type='instance', meter_name='instance_cpu_usage',
            period=300, granularity=360, aggregate='mean',
        )

        # restore the metric map as it is a static attribute that does not get
        # restored between unit tests!
        helper.METRIC_MAP.update(
            instance_cpu_usage=original_metric_value
        )

    def test_get_host_cpu_usage(self, mock_gnocchi):
        self.helper.get_host_cpu_usage('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_cpu_usage', 600, 'mean',
            300)

    def test_get_host_ram_usage(self, mock_gnocchi):
        self.helper.get_host_ram_usage('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_ram_usage', 600, 'mean',
            300)

    def test_get_host_outlet_temperature(self, mock_gnocchi):
        self.helper.get_host_outlet_temp('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_outlet_temp', 600, 'mean',
            300)

    def test_get_host_inlet_temperature(self, mock_gnocchi):
        self.helper.get_host_inlet_temp('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_inlet_temp', 600, 'mean',
            300)

    def test_get_host_airflow(self, mock_gnocchi):
        self.helper.get_host_airflow('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_airflow', 600, 'mean',
            300)

    def test_get_host_power(self, mock_gnocchi):
        self.helper.get_host_power('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'compute_node', 'host_power', 600, 'mean',
            300)

    def test_get_instance_cpu_usage(self, mock_gnocchi):
        self.helper.get_instance_cpu_usage('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'instance', 'instance_cpu_usage', 600, 'mean',
            300)

    def test_get_instance_memory_usage(self, mock_gnocchi):
        self.helper.get_instance_ram_usage('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'instance', 'instance_ram_usage', 600, 'mean',
            300)

    def test_get_instance_ram_allocated(self, mock_gnocchi):
        self.helper.get_instance_ram_allocated('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'instance', 'instance_ram_allocated', 600, 'mean',
            300)

    def test_get_instance_root_disk_allocated(self, mock_gnocchi):
        self.helper.get_instance_root_disk_size('compute1', 600, 'mean', 300)
        self.mock_aggregation.assert_called_once_with(
            'compute1', 'instance', 'instance_root_disk_size', 600, 'mean',
            300)

    def test_gnocchi_check_availability(self, mock_gnocchi):
        gnocchi = mock.MagicMock()
        gnocchi.status.get.return_value = True
        mock_gnocchi.return_value = gnocchi
        helper = gnocchi_helper.GnocchiHelper()
        result = helper.check_availability()
        self.assertEqual('available', result)

    def test_gnocchi_check_availability_with_failure(self, mock_gnocchi):
        cfg.CONF.set_override("query_max_retries", 1,
                              group='watcher_datasources')
        gnocchi = mock.MagicMock()
        gnocchi.status.get.side_effect = Exception()
        mock_gnocchi.return_value = gnocchi
        helper = gnocchi_helper.GnocchiHelper()

        self.assertEqual('not available', helper.check_availability())

    def test_gnocchi_list_metrics(self, mock_gnocchi):
        gnocchi = mock.MagicMock()
        metrics = [{"name": "metric1"}, {"name": "metric2"}]
        expected_metrics = set(["metric1", "metric2"])
        gnocchi.metric.list.return_value = metrics
        mock_gnocchi.return_value = gnocchi
        helper = gnocchi_helper.GnocchiHelper()
        result = helper.list_metrics()
        self.assertEqual(expected_metrics, result)

    def test_gnocchi_list_metrics_with_failure(self, mock_gnocchi):
        cfg.CONF.set_override("query_max_retries", 1,
                              group='watcher_datasources')
        gnocchi = mock.MagicMock()
        gnocchi.metric.list.side_effect = Exception()
        mock_gnocchi.return_value = gnocchi
        helper = gnocchi_helper.GnocchiHelper()
        self.assertFalse(helper.list_metrics())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from watcher.common.policies import base

AUDIT_TEMPLATE = 'audit_template:%s'

rules = [
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'create',
        check_str=base.RULE_ADMIN_API,
        description='Create an audit template.',
        operations=[
            {
                'path': '/v1/audit_templates',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'delete',
        check_str=base.RULE_ADMIN_API,
        description='Delete an audit template.',
        operations=[
            {
                'path': '/v1/audit_templates/{audit_template_uuid}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'detail',
        check_str=base.RULE_ADMIN_API,
        description='Retrieve a list of audit templates with details.',
        operations=[
            {
                'path': '/v1/audit_templates/detail',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'get',
        check_str=base.RULE_ADMIN_API,
        description='Get an audit template.',
        operations=[
            {
                'path': '/v1/audit_templates/{audit_template_uuid}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'get_all',
        check_str=base.RULE_ADMIN_API,
        description='Get a list of all audit templates.',
        operations=[
            {
                'path': '/v1/audit_templates',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=AUDIT_TEMPLATE % 'update',
        check_str=base.RULE_ADMIN_API,
        description='Update an audit template.',
        operations=[
            {
                'path': '/v1/audit_templates/{audit_template_uuid}',
                'method': 'PATCH'
            }
        ]
    )
]


def list_rules():
    return rules



# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_log import log

from watcher.common import exception
from watcher.common import utils
from watcher.decision_engine.model import element
from watcher.decision_engine.planner import base
from watcher import objects

LOG = log.getLogger(__name__)


class NodeResourceConsolidationPlanner(base.BasePlanner):
    """Node Resource Consolidation planner implementation

    This implementation preserves the original order of actions in the
    solution and try to parallelize actions which have the same action type.

    *Limitations*

    - This is a proof of concept that is not meant to be used in production
    """

    def create_action(self,
                      action_plan_id,
                      action_type,
                      input_parameters=None):
        uuid = utils.generate_uuid()
        action = {
            'uuid': uuid,
            'action_plan_id': int(action_plan_id),
            'action_type': action_type,
            'input_parameters': input_parameters,
            'state': objects.action.State.PENDING,
            'parents': None
        }

        return action

    def schedule(self, context, audit_id, solution):
        LOG.debug('Creating an action plan for the audit uuid: %s', audit_id)
        action_plan = self._create_action_plan(context, audit_id, solution)

        actions = list(solution.actions)
        if len(actions) == 0:
            LOG.warning("The action plan is empty")
            action_plan.state = objects.action_plan.State.SUCCEEDED
            action_plan.save()
            return action_plan

        node_disabled_actions = []
        node_enabled_actions = []
        node_migrate_actions = {}
        for action in actions:
            action_type = action.get('action_type')
            parameters = action.get('input_parameters')
            json_action = self.create_action(
                action_plan_id=action_plan.id,
                action_type=action_type,
                input_parameters=parameters)
            # classing actions
            if action_type == 'change_nova_service_state':
                if parameters.get('state') == (
                        element.ServiceState.DISABLED.value):
                    node_disabled_actions.append(json_action)
                else:
                    node_enabled_actions.append(json_action)
            elif action_type == 'migrate':
                source_node = parameters.get('source_node')
                if source_node in node_migrate_actions:
                    node_migrate_actions[source_node].append(json_action)
                else:
                    node_migrate_actions[source_node] = [json_action]
            else:
                raise exception.UnsupportedActionType(
                    action_type=action.get("action_type"))

        # creating actions
        mig_parents = []
        for action in node_disabled_actions:
            mig_parents.append(action['uuid'])
            self._create_action(context, action)

        enabled_parents = []
        for actions in node_migrate_actions.values():
            enabled_parents.append(actions[-1].get('uuid'))
            pre_action_uuid = []
            for action in actions:
                action['parents'] = mig_parents + pre_action_uuid
                pre_action_uuid = [action['uuid']]
                self._create_action(context, action)

        for action in node_enabled_actions:
            action['parents'] = enabled_parents
            self._create_action(context, action)

        self._create_efficacy_indicators(
            context, action_plan.id, solution.efficacy_indicators)

        return action_plan

    def _create_action_plan(self, context, audit_id, solution):
        strategy = objects.Strategy.get_by_name(
            context, solution.strategy.name)

        action_plan_dict = {
            'uuid': utils.generate_uuid(),
            'audit_id': audit_id,
            'strategy_id': strategy.id,
            'state': objects.action_plan.State.RECOMMENDED,
            'global_efficacy': solution.global_efficacy,
        }

        new_action_plan = objects.ActionPlan(context, **action_plan_dict)
        new_action_plan.create()

        return new_action_plan

    def _create_efficacy_indicators(self, context, action_plan_id, indicators):
        efficacy_indicators = []
        for indicator in indicators:
            efficacy_indicator_dict = {
                'uuid': utils.generate_uuid(),
                'name': indicator.name,
                'description': indicator.description,
                'unit': indicator.unit,
                'value': indicator.value,
                'action_plan_id': action_plan_id,
            }
            new_efficacy_indicator = objects.EfficacyIndicator(
                context, **efficacy_indicator_dict)
            new_efficacy_indicator.create()

            efficacy_indicators.append(new_efficacy_indicator)
        return efficacy_indicators

    def _create_action(self, context, _action):
        try:
            LOG.debug("Creating the %s in the Watcher database",
                      _action.get("action_type"))

            new_action = objects.Action(context, **_action)
            new_action.create()

            return new_action
        except Exception as exc:
            LOG.exception(exc)
            raise



# -*- encoding: utf-8 -*-
# Copyright 2019 ZTE Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from unittest import mock

from watcher.common import exception
from watcher.common import utils
from watcher.decision_engine.messaging import data_model_endpoint
from watcher.decision_engine.model.collector import manager
from watcher.objects import audit


class TestDataModelEndpoint(unittest.TestCase):
    def setUp(self):
        self.endpoint_instance = data_model_endpoint.DataModelEndpoint('fake')

    @mock.patch.object(audit.Audit, 'get')
    def test_get_audit_scope(self, mock_get):
        mock_get.return_value = mock.Mock(scope='fake_scope')
        audit_uuid = utils.generate_uuid()

        result = self.endpoint_instance.get_audit_scope(
            context=None,
            audit=audit_uuid)
        self.assertEqual('fake_scope', result)

    @mock.patch.object(audit.Audit, 'get_by_name')
    def test_get_audit_scope_with_error_name(self, mock_get_by_name):
        mock_get_by_name.side_effect = exception.AuditNotFound()
        audit_name = 'error_audit_name'

        self.assertRaises(
            exception.InvalidIdentity,
            self.endpoint_instance.get_audit_scope,
            context=None,
            audit=audit_name)

    @mock.patch.object(manager, 'CollectorManager', mock.Mock())
    def test_get_data_model_info(self):
        result = self.endpoint_instance.get_data_model_info(context='fake')
        self.assertIn('context', result)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

import iso8601

from watcher.db.sqlalchemy import api as db_api
from watcher import objects
from watcher.tests.db import base
from watcher.tests.db import utils


class TestGoalObject(base.DbTestCase):

    def setUp(self):
        super(TestGoalObject, self).setUp()
        self.fake_goal = utils.get_test_goal(
            created_at=datetime.datetime.utcnow())

    @mock.patch.object(db_api.Connection, 'get_goal_by_id')
    def test_get_by_id(self, mock_get_goal):
        goal_id = self.fake_goal['id']
        mock_get_goal.return_value = self.fake_goal
        goal = objects.Goal.get(self.context, goal_id)
        mock_get_goal.assert_called_once_with(self.context, goal_id)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'get_goal_by_uuid')
    def test_get_by_uuid(self, mock_get_goal):
        uuid = self.fake_goal['uuid']
        mock_get_goal.return_value = self.fake_goal
        goal = objects.Goal.get(self.context, uuid)
        mock_get_goal.assert_called_once_with(self.context, uuid)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'get_goal_by_name')
    def test_get_by_name(self, mock_get_goal):
        name = self.fake_goal['name']
        mock_get_goal.return_value = self.fake_goal
        goal = objects.Goal.get_by_name(self.context, name)
        mock_get_goal.assert_called_once_with(self.context, name)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'get_goal_list')
    def test_list(self, mock_get_list):
        mock_get_list.return_value = [self.fake_goal]
        goals = objects.Goal.list(self.context)
        self.assertEqual(1, mock_get_list.call_count)
        self.assertEqual(1, len(goals))
        self.assertIsInstance(goals[0], objects.Goal)
        self.assertEqual(self.context, goals[0]._context)

    @mock.patch.object(db_api.Connection, 'create_goal')
    def test_create(self, mock_create_goal):
        mock_create_goal.return_value = self.fake_goal
        goal = objects.Goal(self.context, **self.fake_goal)
        goal.create()
        expected_goal = self.fake_goal.copy()
        expected_goal['created_at'] = expected_goal['created_at'].replace(
            tzinfo=iso8601.UTC)
        mock_create_goal.assert_called_once_with(expected_goal)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'destroy_goal')
    @mock.patch.object(db_api.Connection, 'get_goal_by_id')
    def test_destroy(self, mock_get_goal, mock_destroy_goal):
        goal_id = self.fake_goal['id']
        mock_get_goal.return_value = self.fake_goal
        goal = objects.Goal.get_by_id(self.context, goal_id)
        goal.destroy()
        mock_get_goal.assert_called_once_with(
            self.context, goal_id)
        mock_destroy_goal.assert_called_once_with(goal_id)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'update_goal')
    @mock.patch.object(db_api.Connection, 'get_goal_by_uuid')
    def test_save(self, mock_get_goal, mock_update_goal):
        mock_get_goal.return_value = self.fake_goal
        goal_uuid = self.fake_goal['uuid']
        fake_saved_goal = self.fake_goal.copy()
        fake_saved_goal['updated_at'] = datetime.datetime.utcnow()
        mock_update_goal.return_value = fake_saved_goal

        goal = objects.Goal.get_by_uuid(self.context, goal_uuid)
        goal.display_name = 'DUMMY'
        goal.save()

        mock_get_goal.assert_called_once_with(self.context, goal_uuid)
        mock_update_goal.assert_called_once_with(
            goal_uuid, {'display_name': 'DUMMY'})
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'get_goal_by_uuid')
    def test_refresh(self, mock_get_goal):
        fake_goal2 = utils.get_test_goal(name="BALANCE_LOAD")
        returns = [self.fake_goal, fake_goal2]
        mock_get_goal.side_effect = returns
        uuid = self.fake_goal['uuid']
        expected = [mock.call(self.context, uuid),
                    mock.call(self.context, uuid)]
        goal = objects.Goal.get(self.context, uuid)
        self.assertEqual("TEST", goal.name)
        goal.refresh()
        self.assertEqual("BALANCE_LOAD", goal.name)
        self.assertEqual(expected, mock_get_goal.call_args_list)
        self.assertEqual(self.context, goal._context)

    @mock.patch.object(db_api.Connection, 'soft_delete_goal')
    @mock.patch.object(db_api.Connection, 'get_goal_by_uuid')
    def test_soft_delete(self, mock_get_goal, mock_soft_delete_goal):
        mock_get_goal.return_value = self.fake_goal
        fake_deleted_goal = self.fake_goal.copy()
        fake_deleted_goal['deleted_at'] = datetime.datetime.utcnow()
        mock_soft_delete_goal.return_value = fake_deleted_goal

        expected_goal = fake_deleted_goal.copy()
        expected_goal['created_at'] = expected_goal['created_at'].replace(
            tzinfo=iso8601.UTC)
        expected_goal['deleted_at'] = expected_goal['deleted_at'].replace(
            tzinfo=iso8601.UTC)

        uuid = self.fake_goal['uuid']
        goal = objects.Goal.get_by_uuid(self.context, uuid)
        goal.soft_delete()
        mock_get_goal.assert_called_once_with(self.context, uuid)
        mock_soft_delete_goal.assert_called_once_with(uuid)
        self.assertEqual(self.context, goal._context)
        self.assertEqual(expected_goal, goal.as_dict())



# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Version 1 of the Watcher API

NOTE: IN PROGRESS AND NOT FULLY IMPLEMENTED.
"""

import datetime

import pecan
from pecan import rest
from webob import exc
import wsme
from wsme import types as wtypes
import wsmeext.pecan as wsme_pecan

from watcher.api.controllers import base
from watcher.api.controllers import link
from watcher.api.controllers.v1 import action
from watcher.api.controllers.v1 import action_plan
from watcher.api.controllers.v1 import audit
from watcher.api.controllers.v1 import audit_template
from watcher.api.controllers.v1 import data_model
from watcher.api.controllers.v1 import goal
from watcher.api.controllers.v1 import scoring_engine
from watcher.api.controllers.v1 import service
from watcher.api.controllers.v1 import strategy
from watcher.api.controllers.v1 import utils
from watcher.api.controllers.v1 import versions
from watcher.api.controllers.v1 import webhooks


def min_version():
    return base.Version(
        {base.Version.string: ' '.join([versions.service_type_string(),
                                        versions.min_version_string()])},
        versions.min_version_string(), versions.max_version_string())


def max_version():
    return base.Version(
        {base.Version.string: ' '.join([versions.service_type_string(),
                                        versions.max_version_string()])},
        versions.min_version_string(), versions.max_version_string())


class APIBase(wtypes.Base):

    created_at = wsme.wsattr(datetime.datetime, readonly=True)
    """The time in UTC at which the object is created"""

    updated_at = wsme.wsattr(datetime.datetime, readonly=True)
    """The time in UTC at which the object is updated"""

    deleted_at = wsme.wsattr(datetime.datetime, readonly=True)
    """The time in UTC at which the object is deleted"""

    def as_dict(self):
        """Render this object as a dict of its fields."""
        return dict((k, getattr(self, k))
                    for k in self.fields
                    if hasattr(self, k) and
                    getattr(self, k) != wsme.Unset)

    def unset_fields_except(self, except_list=None):
        """Unset fields so they don't appear in the message body.

        :param except_list: A list of fields that won't be touched.

        """
        if except_list is None:
            except_list = []

        for k in self.as_dict():
            if k not in except_list:
                setattr(self, k, wsme.Unset)


class MediaType(APIBase):
    """A media type representation."""

    base = wtypes.text
    type = wtypes.text

    def __init__(self, base, type):
        self.base = base
        self.type = type


class V1(APIBase):
    """The representation of the version 1 of the API."""

    id = wtypes.text
    """The ID of the version, also acts as the release number"""

    media_types = [MediaType]
    """An array of supcontainersed media types for this version"""

    audit_templates = [link.Link]
    """Links to the audit templates resource"""

    audits = [link.Link]
    """Links to the audits resource"""

    data_model = [link.Link]
    """Links to the data model resource"""

    actions = [link.Link]
    """Links to the actions resource"""

    action_plans = [link.Link]
    """Links to the action plans resource"""

    scoring_engines = [link.Link]
    """Links to the Scoring Engines resource"""

    services = [link.Link]
    """Links to the services resource"""

    webhooks = [link.Link]
    """Links to the webhooks resource"""

    links = [link.Link]
    """Links that point to a specific URL for this version and documentation"""

    @staticmethod
    def convert():
        v1 = V1()
        v1.id = "v1"
        base_url = pecan.request.application_url
        v1.links = [link.Link.make_link('self', base_url,
                                        'v1', '', bookmark=True),
                    link.Link.make_link('describedby',
                                        'http://docs.openstack.org',
                                        'developer/watcher/dev',
                                        'api-spec-v1.html',
                                        bookmark=True, type='text/html')
                    ]
        v1.media_types = [MediaType('application/json',
                          'application/vnd.openstack.watcher.v1+json')]
        v1.audit_templates = [link.Link.make_link('self',
                                                  base_url,
                                                  'audit_templates', ''),
                              link.Link.make_link('bookmark',
                                                  base_url,
                                                  'audit_templates', '',
                                                  bookmark=True)
                              ]
        v1.audits = [link.Link.make_link('self', base_url,
                                         'audits', ''),
                     link.Link.make_link('bookmark',
                                         base_url,
                                         'audits', '',
                                         bookmark=True)
                     ]
        if utils.allow_list_datamodel():
            v1.data_model = [link.Link.make_link('self', base_url,
                                                 'data_model', ''),
                             link.Link.make_link('bookmark',
                                                 base_url,
                                                 'data_model', '',
                                                 bookmark=True)
                             ]
        v1.actions = [link.Link.make_link('self', base_url,
                                          'actions', ''),
                      link.Link.make_link('bookmark',
                                          base_url,
                                          'actions', '',
                                          bookmark=True)
                      ]
        v1.action_plans = [link.Link.make_link(
            'self', base_url, 'action_plans', ''),
            link.Link.make_link('bookmark',
                                base_url,
                                'action_plans', '',
                                bookmark=True)
            ]

        v1.scoring_engines = [link.Link.make_link(
            'self', base_url, 'scoring_engines', ''),
            link.Link.make_link('bookmark',
                                base_url,
                                'scoring_engines', '',
                                bookmark=True)
            ]

        v1.services = [link.Link.make_link(
            'self', base_url, 'services', ''),
            link.Link.make_link('bookmark',
                                base_url,
                                'services', '',
                                bookmark=True)
            ]
        if utils.allow_webhook_api():
            v1.webhooks = [link.Link.make_link(
                'self', base_url, 'webhooks', ''),
                link.Link.make_link('bookmark',
                                    base_url,
                                    'webhooks', '',
                                    bookmark=True)
                ]
        return v1


class Controller(rest.RestController):
    """Version 1 API controller root."""

    audits = audit.AuditsController()
    audit_templates = audit_template.AuditTemplatesController()
    actions = action.ActionsController()
    action_plans = action_plan.ActionPlansController()
    goals = goal.GoalsController()
    scoring_engines = scoring_engine.ScoringEngineController()
    services = service.ServicesController()
    strategies = strategy.StrategiesController()
    data_model = data_model.DataModelController()
    webhooks = webhooks.WebhookController()

    @wsme_pecan.wsexpose(V1)
    def get(self):
        # NOTE: The reason why convert() it's being called for every
        #       request is because we need to get the host url from
        #       the request object to make the links.
        return V1.convert()

    def _check_version(self, version, headers=None):
        if headers is None:
            headers = {}
        # ensure that major version in the URL matches the header
        if version.major != versions.BASE_VERSION:
            raise exc.HTTPNotAcceptable(
                "Mutually exclusive versions requested. Version %(ver)s "
                "requested but not supported by this service. The supported "
                "version range is: [%(min)s, %(max)s]." %
                {'ver': version, 'min': versions.min_version_string(),
                 'max': versions.max_version_string()},
                headers=headers)
        # ensure the minor version is within the supported range
        if version < min_version() or version > max_version():
            raise exc.HTTPNotAcceptable(
                "Version %(ver)s was requested but the minor version is not "
                "supported by this service. The supported version range is: "
                "[%(min)s, %(max)s]." %
                {'ver': version, 'min': versions.min_version_string(),
                 'max': versions.max_version_string()},
                headers=headers)

    @pecan.expose()
    def _route(self, args, request=None):
        v = base.Version(pecan.request.headers, versions.min_version_string(),
                         versions.max_version_string())

        # The Vary header is used as a hint to caching proxies and user agents
        # that the response is also dependent on the OpenStack-API-Version and
        # not just the body and query parameters. See RFC 7231 for details.
        pecan.response.headers['Vary'] = base.Version.string

        # Always set the min and max headers
        pecan.response.headers[base.Version.min_string] = (
            versions.min_version_string())
        pecan.response.headers[base.Version.max_string] = (
            versions.max_version_string())

        # assert that requested version is supported
        self._check_version(v, pecan.response.headers)
        pecan.response.headers[base.Version.string] = (
            ' '.join([versions.service_type_string(), str(v)]))
        pecan.request.version = v

        return super(Controller, self)._route(args, request)


__all__ = ("Controller", )



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from taskflow import engines
from taskflow import exceptions as tf_exception
from taskflow.patterns import graph_flow as gf
from taskflow import task as flow_task

from watcher.applier.workflow_engine import base
from watcher.common import exception
from watcher import conf
from watcher import objects

CONF = conf.CONF

LOG = log.getLogger(__name__)


class DefaultWorkFlowEngine(base.BaseWorkFlowEngine):
    """Taskflow as a workflow engine for Watcher

    Full documentation on taskflow at
    https://docs.openstack.org/taskflow/latest
    """

    def decider(self, history):
        # decider – A callback function that will be expected to
        # decide at runtime whether v should be allowed to execute
        # (or whether the execution of v should be ignored,
        # and therefore not executed). It is expected to take as single
        # keyword argument history which will be the execution results of
        # all u decidable links that have v as a target. It is expected
        # to return a single boolean
        # (True to allow v execution or False to not).
        LOG.info("decider history: %s", history)
        if history and self.execution_rule == 'ANY':
            return not list(history.values())[0]
        else:
            return True

    @classmethod
    def get_config_opts(cls):
        return [
            cfg.IntOpt(
                'max_workers',
                default=processutils.get_worker_count(),
                min=1,
                required=True,
                help='Number of workers for taskflow engine '
                     'to execute actions.'),
            cfg.DictOpt(
                'action_execution_rule',
                default={},
                help='The execution rule for linked actions,'
                     'the key is strategy name and '
                     'value ALWAYS means all actions will be executed,'
                     'value ANY means if previous action executes '
                     'success, the next action will be ignored.'
                     'None means ALWAYS.')
            ]

    def get_execution_rule(self, actions):
        if actions:
            actionplan_object = objects.ActionPlan.get_by_id(
                self.context, actions[0].action_plan_id)
            strategy_object = objects.Strategy.get_by_id(
                self.context, actionplan_object.strategy_id)
            return self.config.action_execution_rule.get(
                strategy_object.name)

    def execute(self, actions):
        try:
            # NOTE(jed) We want to have a strong separation of concern
            # between the Watcher planner and the Watcher Applier in order
            # to us the possibility to support several workflow engine.
            # We want to provide the 'taskflow' engine by
            # default although we still want to leave the possibility for
            # the users to change it.
            # The current implementation uses graph with linked actions.
            # todo(jed) add oslo conf for retry and name
            self.execution_rule = self.get_execution_rule(actions)
            flow = gf.Flow("watcher_flow")
            actions_uuid = {}
            for a in actions:
                task = TaskFlowActionContainer(a, self)
                flow.add(task)
                actions_uuid[a.uuid] = task

            for a in actions:
                for parent_id in a.parents:
                    flow.link(actions_uuid[parent_id], actions_uuid[a.uuid],
                              decider=self.decider)

            e = engines.load(
                flow, executor='greenthreaded', engine='parallel',
                max_workers=self.config.max_workers)
            e.run()

            return flow

        except exception.ActionPlanCancelled:
            raise

        except tf_exception.WrappedFailure as e:
            if e.check("watcher.common.exception.ActionPlanCancelled"):
                raise exception.ActionPlanCancelled
            else:
                raise exception.WorkflowExecutionException(error=e)

        except Exception as e:
            raise exception.WorkflowExecutionException(error=e)


class TaskFlowActionContainer(base.BaseTaskFlowActionContainer):
    def __init__(self, db_action, engine):
        self.name = "action_type:{0} uuid:{1}".format(db_action.action_type,
                                                      db_action.uuid)
        super(TaskFlowActionContainer, self).__init__(self.name,
                                                      db_action,
                                                      engine)

    def do_pre_execute(self):
        db_action = self.engine.notify(self._db_action,
                                       objects.action.State.ONGOING)
        LOG.debug("Pre-condition action: %s", self.name)
        self.action.pre_condition()
        return db_action

    def do_execute(self, *args, **kwargs):
        LOG.debug("Running action: %s", self.name)

        # NOTE:Some actions(such as migrate) will return None when exception
        #      Only when True is returned, the action state is set to SUCCEEDED
        result = self.action.execute()
        if result is True:
            return self.engine.notify(self._db_action,
                                      objects.action.State.SUCCEEDED)
        else:
            self.engine.notify(self._db_action,
                               objects.action.State.FAILED)
            raise exception.ActionExecutionFailure(
                action_id=self._db_action.uuid)

    def do_post_execute(self):
        LOG.debug("Post-condition action: %s", self.name)
        self.action.post_condition()

    def do_revert(self, *args, **kwargs):
        # NOTE: Not rollback action plan
        if not CONF.watcher_applier.rollback_when_actionplan_failed:
            LOG.info("Failed actionplan rollback option is turned off, and "
                     "the following action will be skipped: %s", self.name)
            return

        LOG.warning("Revert action: %s", self.name)
        try:
            # TODO(jed): do we need to update the states in case of failure?
            self.action.revert()
        except Exception as e:
            LOG.exception(e)
            LOG.critical("Oops! We need a disaster recover plan.")

    def do_abort(self, *args, **kwargs):
        LOG.warning("Aborting action: %s", self.name)
        try:
            result = self.action.abort()
            if result:
                # Aborted the action.
                return self.engine.notify(self._db_action,
                                          objects.action.State.CANCELLED)
            else:
                return self.engine.notify(self._db_action,
                                          objects.action.State.SUCCEEDED)
        except Exception as e:
            LOG.exception(e)
            return self.engine.notify(self._db_action,
                                      objects.action.State.FAILED)


class TaskFlowNop(flow_task.Task):
    """This class is used in case of the workflow have only one Action.

    We need at least two atoms to create a link.
    """
    def execute(self):
        pass



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import prettytable as ptable

from watcher.applier.loading import default as applier_loader
from watcher.common import utils
from watcher.decision_engine.loading import default as decision_engine_loader

PLUGIN_LOADERS = (
    applier_loader.DefaultActionLoader,
    decision_engine_loader.DefaultPlannerLoader,
    decision_engine_loader.DefaultScoringLoader,
    decision_engine_loader.DefaultScoringContainerLoader,
    decision_engine_loader.DefaultStrategyLoader,
    decision_engine_loader.ClusterDataModelCollectorLoader,
    applier_loader.DefaultWorkFlowEngineLoader,
)


def list_opts():
    """Load config options for all Watcher plugins"""
    plugins_opts = []
    for plugin_loader_cls in PLUGIN_LOADERS:
        plugin_loader = plugin_loader_cls()
        plugins_map = plugin_loader.list_available()

        for plugin_name, plugin_cls in plugins_map.items():
            plugin_opts = plugin_cls.get_config_opts()
            if plugin_opts:
                plugins_opts.append(
                    (plugin_loader.get_entry_name(plugin_name), plugin_opts))

    return plugins_opts


def _show_plugins_ascii_table(rows):
    headers = ["Namespace", "Plugin name", "Import path"]
    table = ptable.PrettyTable(field_names=headers)
    for row in rows:
        table.add_row(row)
    return table.get_string()


def show_plugins():
    rows = []
    for plugin_loader_cls in PLUGIN_LOADERS:
        plugin_loader = plugin_loader_cls()
        plugins_map = plugin_loader.list_available()

        rows += [
            (plugin_loader.get_entry_name(plugin_name),
             plugin_name,
             utils.get_cls_import_path(plugin_cls))
            for plugin_name, plugin_cls in plugins_map.items()]

    return _show_plugins_ascii_table(rows)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import jsonschema

from watcher.applier.actions import base as baction
from watcher.applier.actions import change_nova_service_state
from watcher.common import clients
from watcher.common import nova_helper
from watcher.decision_engine.model import element
from watcher.tests import base


class TestChangeNovaServiceState(base.TestCase):

    def setUp(self):
        super(TestChangeNovaServiceState, self).setUp()

        self.m_osc_cls = mock.Mock()
        self.m_helper_cls = mock.Mock()
        self.m_helper = mock.Mock(spec=nova_helper.NovaHelper)
        self.m_helper_cls.return_value = self.m_helper
        self.m_osc = mock.Mock(spec=clients.OpenStackClients)
        self.m_osc_cls.return_value = self.m_osc

        m_openstack_clients = mock.patch.object(
            clients, "OpenStackClients", self.m_osc_cls)
        m_nova_helper = mock.patch.object(
            nova_helper, "NovaHelper", self.m_helper_cls)

        m_openstack_clients.start()
        m_nova_helper.start()

        self.addCleanup(m_openstack_clients.stop)
        self.addCleanup(m_nova_helper.stop)

        self.input_parameters = {
            "resource_name": "compute-1",
            "state": element.ServiceState.ENABLED.value,
        }
        self.action = change_nova_service_state.ChangeNovaServiceState(
            mock.Mock())
        self.action.input_parameters = self.input_parameters

    def test_parameters_down(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: "compute-1",
            self.action.STATE: element.ServiceState.DISABLED.value}
        self.assertTrue(self.action.validate_parameters())

    def test_parameters_up(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: "compute-1",
            self.action.STATE: element.ServiceState.ENABLED.value}
        self.assertTrue(self.action.validate_parameters())

    def test_parameters_exception_wrong_state(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: "compute-1",
            self.action.STATE: 'error'}
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_resource_id_empty(self):
        self.action.input_parameters = {
            self.action.STATE: element.ServiceState.ENABLED.value,
        }
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_applies_add_extra(self):
        self.action.input_parameters = {"extra": "failed"}
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_change_service_state_pre_condition(self):
        try:
            self.action.pre_condition()
        except Exception as exc:
            self.fail(exc)

    def test_change_service_state_post_condition(self):
        try:
            self.action.post_condition()
        except Exception as exc:
            self.fail(exc)

    def test_execute_change_service_state_with_enable_target(self):
        self.action.execute()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.enable_service_nova_compute.assert_called_once_with(
            "compute-1")

    def test_execute_change_service_state_with_disable_target(self):
        self.action.input_parameters["state"] = (
            element.ServiceState.DISABLED.value)
        self.action.input_parameters["disabled_reason"] = (
            "watcher_disabled")
        self.action.execute()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.disable_service_nova_compute.assert_called_once_with(
            "compute-1", "watcher_disabled")

    def test_revert_change_service_state_with_enable_target(self):
        self.action.input_parameters["disabled_reason"] = (
            "watcher_disabled")
        self.action.revert()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.disable_service_nova_compute.assert_called_once_with(
            "compute-1", "watcher_disabled")

    def test_revert_change_service_state_with_disable_target(self):
        self.action.input_parameters["state"] = (
            element.ServiceState.DISABLED.value)
        self.action.revert()

        self.m_helper_cls.assert_called_once_with(osc=self.m_osc)
        self.m_helper.enable_service_nova_compute.assert_called_once_with(
            "compute-1")



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from watcher.common import exception
# from watcher.common import utils as w_utils
from watcher import objects
from watcher.tests.db import base
from watcher.tests.db import utils


class TestEfficacyIndicatorObject(base.DbTestCase):

    def setUp(self):
        super(TestEfficacyIndicatorObject, self).setUp()
        self.fake_efficacy_indicator = utils.get_test_efficacy_indicator()

    def test_get_by_id(self):
        efficacy_indicator_id = self.fake_efficacy_indicator['id']
        with mock.patch.object(self.dbapi, 'get_efficacy_indicator_by_id',
                               autospec=True) as mock_get_efficacy_indicator:
            mock_get_efficacy_indicator.return_value = (
                self.fake_efficacy_indicator)
            efficacy_indicator = objects.EfficacyIndicator.get(
                self.context, efficacy_indicator_id)
            mock_get_efficacy_indicator.assert_called_once_with(
                self.context, efficacy_indicator_id)
            self.assertEqual(self.context, efficacy_indicator._context)

    def test_get_by_uuid(self):
        uuid = self.fake_efficacy_indicator['uuid']
        with mock.patch.object(self.dbapi, 'get_efficacy_indicator_by_uuid',
                               autospec=True) as mock_get_efficacy_indicator:
            mock_get_efficacy_indicator.return_value = (
                self.fake_efficacy_indicator)
            efficacy_indicator = objects.EfficacyIndicator.get(
                self.context, uuid)
            mock_get_efficacy_indicator.assert_called_once_with(
                self.context, uuid)
            self.assertEqual(self.context, efficacy_indicator._context)

    def test_get_bad_id_and_uuid(self):
        self.assertRaises(
            exception.InvalidIdentity,
            objects.EfficacyIndicator.get, self.context, 'not-a-uuid')

    def test_list(self):
        with mock.patch.object(self.dbapi, 'get_efficacy_indicator_list',
                               autospec=True) as mock_get_list:
            mock_get_list.return_value = [self.fake_efficacy_indicator]
            efficacy_indicators = objects.EfficacyIndicator.list(self.context)
            self.assertEqual(1, mock_get_list.call_count)
            self.assertEqual(1, len(efficacy_indicators))
            self.assertIsInstance(
                efficacy_indicators[0], objects.EfficacyIndicator)
            self.assertEqual(self.context, efficacy_indicators[0]._context)

    def test_create(self):
        with mock.patch.object(
            self.dbapi, 'create_efficacy_indicator',
            autospec=True
        ) as mock_create_efficacy_indicator:
            mock_create_efficacy_indicator.return_value = (
                self.fake_efficacy_indicator)
            efficacy_indicator = objects.EfficacyIndicator(
                self.context, **self.fake_efficacy_indicator)

            efficacy_indicator.create()
            mock_create_efficacy_indicator.assert_called_once_with(
                self.fake_efficacy_indicator)
            self.assertEqual(self.context, efficacy_indicator._context)

    def test_destroy(self):
        uuid = self.fake_efficacy_indicator['uuid']
        with mock.patch.object(
            self.dbapi, 'get_efficacy_indicator_by_uuid',
            autospec=True
        ) as mock_get_efficacy_indicator:
            mock_get_efficacy_indicator.return_value = (
                self.fake_efficacy_indicator)
            with mock.patch.object(
                self.dbapi, 'destroy_efficacy_indicator',
                autospec=True
            ) as mock_destroy_efficacy_indicator:
                efficacy_indicator = objects.EfficacyIndicator.get_by_uuid(
                    self.context, uuid)
                efficacy_indicator.destroy()
                mock_get_efficacy_indicator.assert_called_once_with(
                    self.context, uuid)
                mock_destroy_efficacy_indicator.assert_called_once_with(uuid)
                self.assertEqual(self.context, efficacy_indicator._context)

    def test_save(self):
        uuid = self.fake_efficacy_indicator['uuid']
        with mock.patch.object(
            self.dbapi, 'get_efficacy_indicator_by_uuid',
            autospec=True
        ) as mock_get_efficacy_indicator:
            mock_get_efficacy_indicator.return_value = (
                self.fake_efficacy_indicator)
            with mock.patch.object(
                self.dbapi, 'update_efficacy_indicator',
                autospec=True
            ) as mock_update_efficacy_indicator:
                efficacy_indicator = objects.EfficacyIndicator.get_by_uuid(
                    self.context, uuid)
                efficacy_indicator.description = 'Indicator Description'
                efficacy_indicator.save()

                mock_get_efficacy_indicator.assert_called_once_with(
                    self.context, uuid)
                mock_update_efficacy_indicator.assert_called_once_with(
                    uuid, {'description': 'Indicator Description'})
                self.assertEqual(self.context, efficacy_indicator._context)

    def test_refresh(self):
        uuid = self.fake_efficacy_indicator['uuid']
        returns = [dict(self.fake_efficacy_indicator,
                        description="first description"),
                   dict(self.fake_efficacy_indicator,
                        description="second description")]
        expected = [mock.call(self.context, uuid),
                    mock.call(self.context, uuid)]
        with mock.patch.object(self.dbapi, 'get_efficacy_indicator_by_uuid',
                               side_effect=returns,
                               autospec=True) as mock_get_efficacy_indicator:
            efficacy_indicator = objects.EfficacyIndicator.get(
                self.context, uuid)
            self.assertEqual(
                "first description", efficacy_indicator.description)
            efficacy_indicator.refresh()
            self.assertEqual(
                "second description", efficacy_indicator.description)
            self.assertEqual(
                expected, mock_get_efficacy_indicator.call_args_list)
            self.assertEqual(self.context, efficacy_indicator._context)



# -*- encoding: utf-8 -*-
#
# Copyright © 2016 Servionica
##
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class ServiceManager(object, metaclass=abc.ABCMeta):

    @abc.abstractproperty
    def service_name(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def api_version(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def publisher_id(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def conductor_topic(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def notification_topics(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def conductor_endpoints(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def notification_endpoints(self):
        raise NotImplementedError()



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from watcher.applier.loading import default
from watcher.common import utils
from watcher.decision_engine.strategy import strategies
from watcher.tests.decision_engine.strategy.strategies.test_base \
    import TestBaseStrategy


class TestDummyWithScorer(TestBaseStrategy):

    def setUp(self):
        super(TestDummyWithScorer, self).setUp()
        self.strategy = strategies.DummyWithScorer(config=mock.Mock())

    def test_dummy_with_scorer(self):
        dummy = strategies.DummyWithScorer(config=mock.Mock())
        dummy.input_parameters = utils.Struct()
        dummy.input_parameters.update({'param1': 4.0, 'param2': 'Hi'})
        solution = dummy.execute()
        self.assertEqual(4, len(solution.actions))

    def test_check_parameters(self):
        model = self.fake_c_cluster.generate_scenario_3_with_2_nodes()
        self.m_c_model.return_value = model
        self.strategy.input_parameters = utils.Struct()
        self.strategy.input_parameters.update({'param1': 4.0, 'param2': 'Hi'})
        solution = self.strategy.execute()
        loader = default.DefaultActionLoader()
        for action in solution.actions:
            loaded_action = loader.load(action['action_type'])
            loaded_action.input_parameters = action['input_parameters']
            loaded_action.validate_parameters()



# -*- encoding: utf-8 -*-
# Copyright (c) 2019 European Organization for Nuclear Research (CERN)
#
# Authors: Corne Lukken <info@dantalion.nl>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

grafana_client = cfg.OptGroup(name='grafana_client',
                              title='Configuration Options for Grafana',
                              help="See https://docs.openstack.org/watcher/lat"
                                   "est/datasources/grafana.html for details "
                                   "on how these options are used.")

GRAFANA_CLIENT_OPTS = [
    # TODO(Dantali0n) each individual metric could have its own token.
    # A similar structure to the database_map would solve this.
    cfg.StrOpt('token',
               default=None,
               help="Authentication token to gain access"),
    # TODO(Dantali0n) each individual metric could have its own base url.
    # A similar structure to the database_map would solve this.
    cfg.StrOpt('base_url',
               default=None,
               help="First part of the url (including https:// or http://) up "
                    "until project id part. "
                    "Example: https://secure.org/api/datasource/proxy/"),
    cfg.DictOpt('project_id_map',
                default={
                    'host_cpu_usage': None,
                    'host_ram_usage': None,
                    'host_outlet_temp': None,
                    'host_inlet_temp': None,
                    'host_airflow': None,
                    'host_power': None,
                    'instance_cpu_usage': None,
                    'instance_ram_usage': None,
                    'instance_ram_allocated': None,
                    'instance_l3_cache_usage': None,
                    'instance_root_disk_size': None,
                },
                help="Mapping of datasource metrics to grafana project ids. "
                     "Dictionary values should be positive integers. "
                     "Example: 7465"),
    cfg.DictOpt('database_map',
                default={
                    'host_cpu_usage': None,
                    'host_ram_usage': None,
                    'host_outlet_temp': None,
                    'host_inlet_temp': None,
                    'host_airflow': None,
                    'host_power': None,
                    'instance_cpu_usage': None,
                    'instance_ram_usage': None,
                    'instance_ram_allocated': None,
                    'instance_l3_cache_usage': None,
                    'instance_root_disk_size': None,
                },
                help="Mapping of datasource metrics to grafana databases. "
                     "Values should be strings. Example: influx_production"),
    cfg.DictOpt('attribute_map',
                default={
                    'host_cpu_usage': None,
                    'host_ram_usage': None,
                    'host_outlet_temp': None,
                    'host_inlet_temp': None,
                    'host_airflow': None,
                    'host_power': None,
                    'instance_cpu_usage': None,
                    'instance_ram_usage': None,
                    'instance_ram_allocated': None,
                    'instance_l3_cache_usage': None,
                    'instance_root_disk_size': None,
                },
                help="Mapping of datasource metrics to resource attributes. "
                     "For a complete list of available attributes see "
                     "https://docs.openstack.org/watcher/latest/datasources/gr"
                     "afana.html#attribute "
                     "Values should be strings. Example: hostname"),
    cfg.DictOpt('translator_map',
                default={
                    'host_cpu_usage': None,
                    'host_ram_usage': None,
                    'host_outlet_temp': None,
                    'host_inlet_temp': None,
                    'host_airflow': None,
                    'host_power': None,
                    'instance_cpu_usage': None,
                    'instance_ram_usage': None,
                    'instance_ram_allocated': None,
                    'instance_l3_cache_usage': None,
                    'instance_root_disk_size': None,
                },
                help="Mapping of datasource metrics to grafana translators. "
                     "Values should be strings. Example: influxdb"),
    cfg.DictOpt('query_map',
                # {0} = aggregate
                # {1} = attribute
                # {2} = period
                # {3} = granularity
                # {4} = { influxdb: retention_period, }
                default={
                    'host_cpu_usage': None,
                    'host_ram_usage': None,
                    'host_outlet_temp': None,
                    'host_inlet_temp': None,
                    'host_airflow': None,
                    'host_power': None,
                    'instance_cpu_usage': None,
                    'instance_ram_usage': None,
                    'instance_ram_allocated': None,
                    'instance_l3_cache_usage': None,
                    'instance_root_disk_size': None,
                },
                help="Mapping of datasource metrics to grafana queries. "
                     "Values should be strings for which the .format method "
                     "will transform it. The transformation offers five "
                     "parameters to the query labeled {0} to {4}. {0} will be "
                     "replaced with the aggregate, {1} with the resource "
                     "attribute, {2} with the period, {3} with the "
                     "granularity and {4} with translator specifics for "
                     "InfluxDB this will be the retention period. "
                     "These queries will need to be constructed using tools "
                     "such as Postman. Example: SELECT cpu FROM {4}."
                     "cpu_percent WHERE host == '{1}' AND time > now()-{2}s")]


def register_opts(conf):
    conf.register_group(grafana_client)
    conf.register_opts(GRAFANA_CLIENT_OPTS, group=grafana_client)


def list_opts():
    return [(grafana_client, GRAFANA_CLIENT_OPTS)]



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
# Copyright (c) 2016 Intel Corp
#
# Authors: Jean-Emile DARTOIS <jean-emile.dartois@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcher.common import exception
from watcher.common import service
from watcher.common import service_manager
from watcher.common import utils

from watcher import conf

CONF = conf.CONF


class ApplierAPI(service.Service):

    def __init__(self):
        super(ApplierAPI, self).__init__(ApplierAPIManager)

    def launch_action_plan(self, context, action_plan_uuid=None):
        if not utils.is_uuid_like(action_plan_uuid):
            raise exception.InvalidUuidOrName(name=action_plan_uuid)

        self.conductor_client.cast(
            context, 'launch_action_plan', action_plan_uuid=action_plan_uuid)


class ApplierAPIManager(service_manager.ServiceManager):

    @property
    def service_name(self):
        return None

    @property
    def api_version(self):
        return '1.0'

    @property
    def publisher_id(self):
        return CONF.watcher_applier.publisher_id

    @property
    def conductor_topic(self):
        return CONF.watcher_applier.conductor_topic

    @property
    def notification_topics(self):
        return []

    @property
    def conductor_endpoints(self):
        return []

    @property
    def notification_endpoints(self):
        return []



# -*- encoding: utf-8 -*-
# Copyright (c) 2016 b<>com
#
# Authors: Vincent FRANCOISE <vincent.francoise@b-com.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from oslo_serialization import jsonutils

from watcher.common import context
from watcher.common import service as watcher_service
from watcher.decision_engine.model.notification import base
from watcher.decision_engine.model.notification import filtering
from watcher.tests import base as base_test
from watcher.tests.decision_engine.model.notification import fake_managers


class DummyManager(fake_managers.FakeManager):

    @property
    def notification_endpoints(self):
        return [DummyNotification(self.fake_cdmc)]


class DummyNotification(base.NotificationEndpoint):

    @property
    def filter_rule(self):
        return filtering.NotificationFilter(
            publisher_id=r'.*',
            event_type=r'compute.dummy',
            payload={'data': {'nested': r'^T.*'}},
        )

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        pass


class NotificationTestCase(base_test.TestCase):

    def load_message(self, filename):
        cwd = os.path.abspath(os.path.dirname(__file__))
        data_folder = os.path.join(cwd, "data")

        with open(os.path.join(data_folder, filename), 'rb') as json_file:
            json_data = jsonutils.load(json_file)

        return json_data


class TestReceiveNotifications(NotificationTestCase):

    def setUp(self):
        super(TestReceiveNotifications, self).setUp()

        p_from_dict = mock.patch.object(context.RequestContext, 'from_dict')
        m_from_dict = p_from_dict.start()
        m_from_dict.return_value = self.context
        self.addCleanup(p_from_dict.stop)

    @mock.patch.object(watcher_service.ServiceHeartbeat, 'send_beat')
    @mock.patch.object(DummyNotification, 'info')
    def test_receive_dummy_notification(self, m_info, m_heartbeat):
        message = {
            'publisher_id': 'nova-compute',
            'event_type': 'compute.dummy',
            'payload': {'data': {'nested': 'TEST'}},
            'priority': 'INFO',
        }
        de_service = watcher_service.Service(DummyManager)
        incoming = mock.Mock(ctxt=self.context.to_dict(), message=message)

        de_service.notification_handler.dispatcher.dispatch(incoming)

        m_info.assert_called_once_with(
            self.context, 'nova-compute', 'compute.dummy',
            {'data': {'nested': 'TEST'}},
            {'message_id': None, 'timestamp': None})

    @mock.patch.object(watcher_service.ServiceHeartbeat, 'send_beat')
    @mock.patch.object(DummyNotification, 'info')
    def test_skip_unwanted_notification(self, m_info, m_heartbeat):
        message = {
            'publisher_id': 'nova-compute',
            'event_type': 'compute.dummy',
            'payload': {'data': {'nested': 'unwanted'}},
            'priority': 'INFO',
        }
        de_service = watcher_service.Service(DummyManager)
        incoming = mock.Mock(ctxt=self.context.to_dict(), message=message)

        de_service.notification_handler.dispatcher.dispatch(incoming)

        self.assertEqual(0, m_info.call_count)
